# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Parts of the code have been taken from https://github.com/facebookresearch/fastMRI

from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import h5py
import numpy as np
import torch

__all__ = [
    "add_coil_dim_if_singlecoil",
    "apply_mask",
    "batched_mask_center",
    "center_crop",
    "center_crop_to_smallest",
    "check_stacked_complex",
    "coil_combination_method",
    "complex_abs",
    "complex_abs_sq",
    "complex_center_crop",
    "complex_conj",
    "complex_mul",
    "crop_to_acs",
    "expand_op",
    "is_none",
    "mask_center",
    "normalize_inplace",
    "parse_list_and_keep_last",
    "reshape_fortran",
    "rnn_weights_init",
    "rss",
    "rss_complex",
    "save_predictions",
    "sense",
    "to_tensor",
    "unnormalize",
    "zero_nan_inf",
]


def add_coil_dim_if_singlecoil(x: torch.tensor, dim: int = 0) -> torch.tensor:
    """
    Add dummy coil dimension if single coil data.

    Parameters
    ----------
    x : torch.tensor
        The input data.
    dim : int
        The dimension to add coil dimension. Default is ``0``.

    Returns
    -------
    torch.tensor
        The input data with coil dimension added if single coil.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.parts.utils import add_coil_dim_if_singlecoil
    >>> data = torch.rand(10, 10)
    >>> data.shape
    (10, 10)
    >>> add_coil_dim_if_singlecoil(data).shape
    (1, 10, 10)
    >>> add_coil_dim_if_singlecoil(data, dim=-1).shape
    (10, 10, 1)
    """
    if len(x.shape) >= 4:
        return x
    return torch.unsqueeze(x, dim=dim)


def apply_mask(
    x: torch.Tensor,
    mask_func: Callable,
    seed: Optional[Union[int, Tuple[int, ...]]] = None,
    padding: Optional[Sequence[int]] = None,
    shift: bool = False,
    partial_fourier_percentage: Optional[float] = 0.0,
    center_scale: Optional[float] = 0.02,
    existing_mask: Optional[torch.Tensor] = None,
) -> Tuple[Any, Any, int]:
    """
    Retrospectively accelerate/subsample k-space data by applying a mask to the input data.

    Parameters
    ----------
    x : torch.Tensor
        The input k-space data. This should have at least 3 dimensions, where dimensions -3 and -2 are the spatial
        dimensions, and the final dimension has size 2 (for complex values).
    mask_func : Callable
        A function that takes a shape (tuple of ints) and a random number seed and returns a mask.
    seed : Optional[Union[int, Tuple[int, ...]]], optional
        Seed for the random number generator. Default is ``None``.
    padding : Optional[Sequence[int]], optional
        Padding value to apply for mask. Default is ``None``.
    shift : bool, optional
        Toggle to shift mask when subsampling. Applicable on 2D data. Default is ``False``.
    partial_fourier_percentage : Optional[float], optional
        Percentage of kspace to be dropped. Default is ``0.0``.
    center_scale : Optional[float], optional
        Scale of the center of the mask. Applicable on Gaussian masks. Default is ``0.02``.
    existing_mask : Optional[torch.Tensor], optional
        When given, use this mask instead of generating a new one. Default is ``None``.

    Returns
    -------
    Tuple[Any, Any, int]
        Tuple containing the masked k-space data, the mask, and the acceleration factor.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import apply_mask
    >>> import torch
    >>> data = torch.tensor([[[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]], \
    [[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]]])
    >>> data.shape
    torch.Size([2, 2, 3, 2])
    >>> mask = torch.tensor([[[1., 1., 1.], [1., 1., 1.]], [[1., 1., 1.], [1., 1., 1.]]])
    >>> mask.shape
    torch.Size([2, 2, 3])
    >>> apply_mask(data, mask)
    (tensor([[[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]],
        [[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]]]),
    tensor([[[1., 1., 1.], [1., 1., 1.]], [[1., 1., 1.], [1., 1., 1.]]]),
    6)
    >>> masked_data, subsampling_mask, acceleration_factor = apply_mask(data, mask)
    >>> masked_data.shape
    torch.Size([2, 2, 3, 2])
    >>> subsampling_mask.shape
    torch.Size([2, 2, 3])
    >>> acceleration_factor
    6
    >>> apply_mask(data, mask, padding=[1, 2], shift=True)
    (tensor([[[[0., 0.], [0., 0.], [0., 0.]], [[1., 1.], [2., 2.], [3., 3.]]],
        [[[0., 0.], [0., 0.], [0., 0.]], [[1., 1.], [2., 2.], [3., 3.]]]]),
    tensor([[[0., 0., 0.], [1., 1., 1.]], [[0., 0., 0.], [1., 1., 1.]]]),
    3)
    >>> masked_data, subsampling_mask, acceleration_factor = apply_mask(data, mask, padding=[1, 2], shift=True)
    >>> masked_data.shape
    torch.Size([2, 2, 3, 2])
    >>> subsampling_mask.shape
    torch.Size([2, 2, 3])
    >>> acceleration_factor
    3
    """
    shape = np.array(x.shape)
    shape[:-3] = 1

    if existing_mask is None:
        mask, acc = mask_func(shape, seed, partial_fourier_percentage=partial_fourier_percentage, scale=center_scale)
    else:
        mask = existing_mask
        acc = mask.size / mask.sum()

    mask = mask.to(x.device)

    if padding is not None and (padding[0] > 0 or padding[1] > 0):
        mask[..., : padding[0], :] = 0
        mask[..., padding[1] :, :] = 0  # padding value inclusive on right of zeros

    if shift:
        mask = torch.fft.fftshift(mask, dim=(1, 2))

    masked_x = x * mask + 0.0  # the + 0.0 removes the sign of the zeros

    return masked_x, mask, acc


def batched_mask_center(
    x: torch.Tensor, mask_from: torch.Tensor, mask_to: torch.Tensor, mask_type: str = "2D"
) -> torch.Tensor:
    """
    Initializes a mask with the center filled in. Can operate with different masks for each batch element.

    Parameters
    ----------
    x : torch.Tensor
        The input image or batch of images. This should have at least 3 dimensions, where dimensions -3 and -2 are the
        spatial dimensions, and the final dimension has size 1 (for real values).
    mask_from : torch.Tensor
        Part of center to start filling.
    mask_to : torch.Tensor
        Part of center to end filling.
    mask_type : str, optional
        Type of mask to apply. Can be either ``1D`` or ``2D``. Default is ``2D``.

    Returns
    -------
    torch.Tensor
        The masked image or batch of images with filled center.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import batched_mask_center
    >>> import torch
    >>> data = torch.randn(1, 32, 320, 320)
    >>> batched_mask_center(data, torch.tensor([140]), torch.tensor([180]))
    """
    if mask_from.shape != mask_to.shape:
        raise ValueError("mask_from and mask_to must match shapes.")
    if mask_from.ndim != 1:
        raise ValueError("mask_from and mask_to must have 1 dimension.")
    if mask_from.shape[0] not in (1, x.shape[0]) or x.shape[0] != mask_to.shape[0]:
        raise ValueError("mask_from and mask_to must have batch_size length.")

    if mask_from.shape[0] == 1:
        mask = mask_center(x, int(mask_from), int(mask_to), mask_type=mask_type)
    else:
        mask = torch.zeros_like(x)
        for i, (start, end) in enumerate(zip(mask_from, mask_to)):
            mask[i, :, :, start:end] = x[i, :, :, start:end]

    return mask


def center_crop(x: torch.Tensor, shape: Tuple[int, int]) -> torch.Tensor:
    """
    Apply a center crop to the input complex image or batch of complex images or real image or batch of real images
    without a complex dimension.

    Parameters
    ----------
    x : torch.Tensor
        The input tensor to be center cropped. It should have at least 2 dimensions and the cropping is applied along
        the last two dimensions.
    shape : Tuple[int, int]
        The output shape. The shape should be smaller than the corresponding dimensions of data.

    Returns
    -------
    torch.Tensor
        The center cropped image or batch of images.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import center_crop
    >>> import torch
    >>> data = torch.tensor([[[1+1j, 2+2j, 3+3j], [1+1j, 2+2j, 3+3j]], [[1+1j, 2+2j, 3+3j], [1+1j, 2+2j, 3+3j]]])
    >>> data.shape
    torch.Size([2, 2, 3])
    >>> center_crop(data, (1, 2))
    tensor([[[2.+2.j, 3.+3.j]], [[2.+2.j, 3.+3.j]]])
    >>> center_crop(data, (1, 2)).shape
    torch.Size([2, 1, 2])
    """
    if not (0 < shape[0] <= x.shape[-2] and 0 < shape[1] <= x.shape[-1]):
        raise ValueError("Invalid shapes.")

    w_from = torch.div((x.shape[-2] - shape[0]), 2, rounding_mode="trunc")
    h_from = torch.div((x.shape[-1] - shape[1]), 2, rounding_mode="trunc")
    w_to = w_from + shape[0]
    h_to = h_from + shape[1]

    return x[..., w_from:w_to, h_from:h_to]


def center_crop_to_smallest(
    x: Union[torch.Tensor, np.ndarray], y: Union[torch.Tensor, np.ndarray]
) -> Tuple[Union[torch.Tensor, np.ndarray], Union[torch.Tensor, np.ndarray]]:
    """
    Apply a center crop on the larger image to the size of the smaller.

    The minimum is taken over dim=-1 and dim=-2. If x is smaller than y at dim=-1 and y is smaller than x at dim=-2,
        then the returned dimension will be a mixture of the two.

    Parameters
    ----------
    x : torch.Tensor or np.ndarray
        The first image.
    y : torch.Tensor or np.ndarray
        The second image.

    Returns
    -------
    Tuple[torch.Tensor or np.ndarray, torch.Tensor or np.ndarray]
        Tuple of x and y, cropped to the minimum size.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import center_crop_to_smallest
    >>> import torch
    >>> data1 = torch.tensor([[[1+1j, 2+2j, 3+3j], [1+1j, 2+2j, 3+3j]], [[1+1j, 2+2j, 3+3j], [1+1j, 2+2j, 3+3j]]])
    >>> data2 = torch.tensor([[[1+1j, 2+2j, 3+3j, 4+4j, 5+5j], [1+1j, 2+2j, 3+3j, 4+4j, 5+5j]], \
    [[1+1j, 2+2j, 3+3j, 4+4j, 5+5j], [1+1j, 2+2j, 3+3j, 4+4j, 5+5j], [1+1j, 2+2j, 3+3j, 4+4j, 5+5j]]])
    >>> data1.shape
    torch.Size([2, 2, 3])
    >>> data2.shape
    torch.Size([2, 3, 5])
    >>> center_crop_to_smallest(data1, data2)
    (tensor([[[1+1j, 2+2j, 3+3j], [1+1j, 2+2j, 3+3j]], [[1+1j, 2+2j, 3+3j], [1+1j, 2+2j, 3+3j]]]), \
    tensor([[[1.+1.j, 2.+2.j, 3.+3.j], [1.+1.j, 2.+2.j, 3.+3.j]], \
    [[1.+1.j, 2.+2.j, 3.+3.j], [1.+1.j, 2.+2.j, 3.+3.j]]]))
    >>> center_crop_to_smallest(data1, data2)[0].shape
    torch.Size([2, 2, 3])
    >>> center_crop_to_smallest(data1, data2)[1].shape
    torch.Size([2, 2, 3])
    >>> center_crop_to_smallest(data2, data1)
    (tensor([[[1.+1.j, 2.+2.j, 3.+3.j], [1.+1.j, 2.+2.j, 3.+3.j]], \
    [[1.+1.j, 2.+2.j, 3.+3.j], [1.+1.j, 2.+2.j, 3.+3.j]]]), \
    tensor([[[1+1j, 2+2j, 3+3j], [1+1j, 2+2j, 3+3j]], [[1+1j, 2+2j, 3+3j], [1+1j, 2+2j, 3+3j]]]))
    >>> center_crop_to_smallest(data2, data1)[0].shape
    torch.Size([2, 2, 3])
    >>> center_crop_to_smallest(data2, data1)[1].shape
    torch.Size([2, 2, 3])
    """
    smallest_width = min(x.shape[-1], y.shape[-1])
    smallest_height = min(x.shape[-2], y.shape[-2])
    x = center_crop(x, (smallest_height, smallest_width))
    y = center_crop(y, (smallest_height, smallest_width))
    return x, y


def check_stacked_complex(x: torch.Tensor) -> torch.Tensor:
    """
    Check if tensor is stacked complex (real & imaginary parts stacked along last dim) and convert it to a combined
    complex tensor.

    Parameters
    ----------
    x : torch.Tensor
        Tensor to check.

    Returns
    -------
    torch.Tensor
        Tensor with stacked complex converted to combined complex.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import check_stacked_complex
    >>> import torch
    >>> data = torch.tensor([1+1j, 2+2j, 3+3j])
    >>> data.shape
    torch.Size([3])
    >>> data = torch.view_as_real(data)
    >>> data.shape
    >>> check_stacked_complex(data)
    tensor([1.+1.j, 2.+2.j, 3.+3.j])
    >>> check_stacked_complex(data).shape
    torch.Size([3])
    >>> data = torch.tensor([1+1j, 2+2j, 3+3j])
    >>> data.shape
    torch.Size([3])
    >>> check_stacked_complex(data)
    tensor([1.+1.j, 2.+2.j, 3.+3.j])
    >>> check_stacked_complex(data).shape
    torch.Size([3])
    """
    return torch.view_as_complex(x) if x.shape[-1] == 2 else x


def coil_combination_method(
    x: torch.Tensor, sensitivity_maps: torch.Tensor, method: str = "SENSE", dim: int = 0
) -> torch.Tensor:
    """
    Selects the coil combination method.

    Parameters
    ----------
    x : torch.Tensor
        The tensor to coil-combine.
    sensitivity_maps : torch.Tensor
        The coil sensitivity maps.
    method : str, optional
        The coil combination method to use. Options are ``"SENSE"``, ``"RSS"``, ``"RSS_COMPLEX"``.
        Default is ``"SENSE"``.
    dim : int, optional
        The dimension to coil-combine along. Default is ``0``.

    Returns
    -------
    torch.Tensor
        Coil-combined tensor with the selected method applied.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import coil_combination_method
    >>> import torch
    >>> data = torch.tensor([[[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]], \
    [[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]]])
    >>> data.shape
    torch.Size([2, 2, 3, 2])
    >>> coil_sensitivity_maps = torch.tensor([[[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]], \
    [[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]]])
    >>> coil_sensitivity_maps.shape
    torch.Size([2, 2, 3, 2])
    >>> coil_combination_method(data, coil_sensitivity_maps, method="SENSE")
    tensor([[[2.8284, 2.8284],
        [5.6569, 5.6569],
        [8.4853, 8.4853]],
        [[2.8284, 2.8284],
        [5.6569, 5.6569],
        [8.4853, 8.4853]]])
    >>> coil_combination_method(data, coil_sensitivity_maps, method="SENSE").shape
    torch.Size([2, 3, 2])
    >>> coil_combination_method(data, coil_sensitivity_maps, method="RSS")
    tensor([[[1.4142, 1.4142],
        [2.8284, 2.8284],
        [4.2426, 4.2426]],
        [[1.4142, 1.4142],
        [2.8284, 2.8284],
        [4.2426, 4.2426]]])
    >>> coil_combination_method(data, coil_sensitivity_maps, method="RSS").shape
    torch.Size([2, 3, 2])
    >>> coil_combination_method(data, coil_sensitivity_maps, method="RSS_COMPLEX")
    tensor([[[1.4142, 1.4142],
        [2.8284, 2.8284],
        [4.2426, 4.2426]],
        [[1.4142, 1.4142],
        [2.8284, 2.8284],
        [4.2426, 4.2426]]])
    >>> coil_combination_method(data, coil_sensitivity_maps, method="RSS_COMPLEX").shape
    torch.Size([2, 3, 2])
    """
    if method == "SENSE":
        return sense(x, sensitivity_maps, dim)
    if method == "RSS":
        return rss(x, dim)
    if method == "RSS_COMPLEX":
        return rss_complex(x, dim)
    raise ValueError("Output type not supported.")


def complex_abs(x: torch.Tensor) -> torch.Tensor:
    """
    Compute the absolute value of a complex valued input tensor.

    Parameters
    ----------
    x : torch.Tensor
        Complex tensor. The last dimension must be of size 2.

    Returns
    -------
    torch.Tensor
        Absolute value of complex tensor.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import complex_abs
    >>> import torch
    >>> data = torch.tensor([1+1j, 2+2j, 3+3j])
    >>> complex_abs(data)
    tensor([1.4142, 2.8284, 4.2426])
    """
    if x.shape[-1] != 2:
        if torch.is_complex(x):
            x = torch.view_as_real(x)
        else:
            raise ValueError("Tensor does not have separate complex dim.")
    return (x**2).sum(dim=-1)


def complex_abs_sq(x: torch.Tensor) -> torch.Tensor:
    """
    Compute the squared absolute value of a complex tensor.

    Parameters
    ----------
    x : torch.Tensor
        Complex tensor. The last dimension must be of size 2.

    Returns
    -------
    torch.Tensor
        Squared absolute value of complex tensor.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import complex_abs_sq
    >>> import torch
    >>> data = torch.tensor([1+1j, 2+2j, 3+3j])
    >>> complex_abs_sq(data)
    tensor([2., 8., 18.])
    """
    if x.shape[-1] != 2:
        if torch.is_complex(x):
            x = torch.view_as_real(x)
        else:
            raise ValueError("Tensor does not have separate complex dim.")
    return (x**2).sum(dim=-1).sqrt()


def complex_center_crop(x: torch.Tensor, shape: Tuple[int, int]) -> torch.Tensor:
    """
    Apply a center crop to the input image or batch of complex images.

    Parameters
    ----------
    x : torch.Tensor
        The input tensor to be center cropped. It should have at least 3 dimensions and the cropping is applied along
        the last two dimensions.
    shape : Tuple[int, int]
        The output shape. The shape should be smaller than the corresponding dimensions of data.

    Returns
    -------
    torch.Tensor
        The complex center cropped image or batch of images.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import complex_center_crop
    >>> import torch
    >>> data = torch.tensor([[[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]], \
    [[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]]])
    >>> data.shape
    torch.Size([2, 2, 3, 2])
    >>> complex_center_crop(data, (1, 2))
    tensor([[[[2., 2.]]],
        [[[2., 2.]]]])
    >>> complex_center_crop(data, (1, 2)).shape
    torch.Size([2, 1, 1, 2])
    """
    if not (0 < shape[0] <= x.shape[-3] and 0 < shape[1] <= x.shape[-2]):
        raise ValueError("Invalid shapes.")

    w_from = torch.div((x.shape[-3] - shape[0]), 2, rounding_mode="trunc")
    h_from = torch.div((x.shape[-2] - shape[1]), 2, rounding_mode="trunc")
    w_to = w_from + shape[0]
    h_to = h_from + shape[1]

    return x[..., w_from:w_to, h_from:h_to, :]


def complex_conj(x: torch.Tensor) -> torch.Tensor:
    """
    Complex conjugate.

    This applies the complex conjugate assuming that the input array has the last dimension as the complex dimension.

    Parameters
    ----------
    x : torch.Tensor
        Complex tensor to apply the complex conjugate to. The last dimension must be of size 2.

    Returns
    -------
    torch.Tensor
        Result of complex conjugate.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import complex_conj
    >>> import torch
    >>> data = torch.tensor([1+1j, 2+2j, 3+3j])
    >>> complex_conj(data)
    tensor([1.-1.j, 2.-2.j, 3.-3.j])
    """
    if x.shape[-1] != 2:
        raise ValueError("Tensor does not have separate complex dim.")
    return torch.stack((x[..., 0], -x[..., 1]), dim=-1)


def complex_mul(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    Complex multiplication.

    This multiplies two complex tensors assuming that they are both stored as real arrays with the last dimension
    being the complex dimension.

    Parameters
    ----------
    x : torch.Tensor
        First complex tensor to multiply. The last dimension must be of size 2.
    y : torch.Tensor
        Second complex tensor to multiply. The last dimension must be of size 2.

    Returns
    -------
    torch.Tensor
        Result of complex multiplication.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import complex_mul
    >>> import torch
    >>> datax = torch.tensor([1+1j, 2+2j, 3+3j])
    >>> datay = torch.tensor([4+4j, 5+5j, 6+6j])
    >>> complex_mul(datax, datay)
    tensor([[-7.+20.j],
            [-4.+16.j],
            [-1.+12.j]])
    """
    if not x.shape[-1] == y.shape[-1] == 2:
        raise ValueError("Tensors do not have separate complex dim.")
    re = x[..., 0] * y[..., 0] - x[..., 1] * y[..., 1]
    im = x[..., 0] * y[..., 1] + x[..., 1] * y[..., 0]
    return torch.stack((re, im), dim=-1)


def crop_to_acs(acs_mask: torch.Tensor, kspace: torch.Tensor) -> torch.Tensor:
    r"""Crops k-space to autocalibration region given the acs_mask.

    Parameters
    ----------
    acs_mask : torch.Tensor
        Autocalibration mask of shape (height, width).
    kspace : torch.Tensor
        K-space of shape (coil, height, width, *).

    Returns
    -------
    torch.Tensor
        Cropped k-space of shape (coil, height, width, *), where height and width are the new dimensions derived from
        the acs_mask.
    """
    nonzero_idxs = torch.nonzero(acs_mask)
    x, y = nonzero_idxs[..., 0], nonzero_idxs[..., 1]
    xl, xr = x.min(), x.max()
    yl, yr = y.min(), y.max()
    return kspace[:, xl : xr + 1, yl : yr + 1, :]


def expand_op(x: torch.Tensor, sensitivity_maps: torch.Tensor, dim: int = 1) -> torch.Tensor:
    """
    Expand a coil-combined image to a multi-coil image.

    Parameters
    ----------
    x : torch.Tensor
        The coil-combined image.
    sensitivity_maps : torch.Tensor
        The sensitivity maps.
    dim : int
        The coil dimension to expand. Default is ``1``.

    Returns
    -------
    torch.Tensor
        The multi-coil image.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.parts.utils import expand_op
    >>> data = torch.rand(1, 200, 200, 2)
    >>> sens = torch.rand(1, 30, 200, 200, 2)
    >>> expand_op(data, sens).shape
    (1, 30, 200, 200, 2)
    """
    return torch.unsqueeze(x, dim=dim) * sensitivity_maps


def is_none(x: Union[Any, None]) -> bool:
    """
    Check if input is None or "None".

    Parameters
    ----------
    x : Union[Any, None]
        Input to check.

    Returns
    -------
    bool
        True if x is None or "None", False otherwise.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import is_none
    >>> is_none(None)
    True
    >>> is_none("None")
    True
    """
    return x is None or str(x).lower() == "none" or "none" in str(x).lower()


def mask_center(
    x: torch.Tensor, mask_from: Optional[int], mask_to: Optional[int], mask_type: str = "2D"
) -> torch.Tensor:
    """
    Apply a center crop to the input real image or batch of real images.

    Parameters
    ----------
    x : torch.Tensor
        The input image or batch of images. This should have at least 3 dimensions, where dimensions -3 and -2 are the
        spatial dimensions, and the final dimension has size 1 (for real values).
    mask_from : Optional[int]
        Part of center to start filling.
    mask_to : Optional[int]
        Part of center to end filling.
    mask_type : str, optional
        Type of mask to apply. Can be either ``1D`` or ``2D``. Default is ``2D``.

    Returns
    -------
    torch.Tensor
        The masked image or batch of images with filled center.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import mask_center
    >>> import torch
    >>> data = torch.tensor([[[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]], \
    [[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]]])
    >>> data.shape
    torch.Size([2, 2, 3, 2])
    >>> mask_center(data, 1, 2)
    tensor([[[[0., 0.], [1., 1.], [0., 0.]], [[0., 0.], [1., 1.], [0., 0.]]],
        [[[0., 0.], [1., 1.], [0., 0.]], [[0., 0.], [1., 1.], [0., 0.]]]])
    >>> mask_center(data, 1, 2).shape
    torch.Size([2, 2, 3, 2])
    """
    mask = torch.zeros_like(x)

    if isinstance(mask_from, list):
        mask_from = mask_from[0]

    if isinstance(mask_to, list):
        mask_to = mask_to[0]

    if mask_type == "1D":
        mask[:, :, :, mask_from:mask_to] = x[:, :, :, mask_from:mask_to]
    elif mask_type == "2D":
        mask[:, :, mask_from:mask_to] = x[:, :, mask_from:mask_to]
    else:
        raise ValueError(f"Unknown mask type {mask_type}")

    return mask


def normalize_inplace(x: torch.Tensor, normalization_type: str = "max") -> torch.Tensor:
    """
    Normalize the input data inplace. This is different from the ``Normalizer`` transformation that normalizes the data
    in a non batch-wise manner.

    Parameters
    ----------
    x : torch.Tensor
        The input data.
    normalization_type : str
        The normalization type. Default is ``"max"``.

    Returns
    -------
    torch.Tensor
        The unnormalized data.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.parts.utils import normalize_inplace
    >>> data = torch.rand(1, 200, 200, 2)
    >>> attrs = {"max": 164.4672133, "min": 0.000279681}
    >>> normalize_inplace(data, attrs).shape
    (1, 200, 200, 2)
    """
    if normalization_type == "max":
        return x / torch.max(torch.abs(x))
    if normalization_type == "minmax":
        min_value = torch.min(torch.abs(x))
        return (x - min_value) / (torch.max(torch.abs(x)) - min_value)
    if normalization_type == "mean_std":
        x = x - torch.mean(torch.abs(x))
        return x / torch.std(torch.abs(x))
    if normalization_type == "mean_var":
        x = x - torch.mean(torch.abs(x))
        return x / torch.var(torch.abs(x))
    if normalization_type == "grayscale":
        x = x - torch.min(torch.abs(x))
        x = x / torch.max(torch.abs(x))
        return x * 255.0
    return x


def parse_list_and_keep_last(x: Union[Any, List[Any]]) -> List[Any]:
    """Parse a list of values and keep the last one, until the last value is not a list."""
    if isinstance(x, list):
        while isinstance(x, list):
            x = x[-1]
    return x


def reshape_fortran(x, shape) -> torch.Tensor:
    """
    Reshapes a tensor in Fortran order. Taken from https://stackoverflow.com/a/63964246

    Parameters
    ----------
    x : torch.Tensor
        Input tensor to be reshaped.
    shape : Sequence[int]
        Shape to reshape the tensor to.

    Returns
    -------
    torch.Tensor
        Reshaped tensor.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import reshape_fortran
    >>> import torch
    >>> data = torch.tensor([[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12]]])
    >>> data.shape
    torch.Size([2, 2, 3])
    >>> reshape_fortran(data, (3, 2, 2))
    tensor([[[ 1,  7],
            [ 4, 10]],
            [[ 2,  8],
            [ 5, 11]],
            [[ 3,  9],
            [ 6, 12]]])
    >>> reshape_fortran(data, (3, 2, 2)).shape
    torch.Size([3, 2, 2])
    """
    return x.permute(*reversed(range(len(x.shape)))).reshape(*reversed(shape)).permute(*reversed(range(len(shape))))


def rnn_weights_init(module: torch.nn.Module, std_init_range: float = 0.02, xavier: bool = True):
    r"""Initialize weights in Recurrent Neural Network.

    Parameters
    ----------
    module : torch.nn.Module
        Module to initialize.
    std_init_range : float
        Standard deviation of normal initializer. Default is ``0.02``.
    xavier : bool
        If True, xavier initializer will be used in Linear layers as in [Vaswani2017]_. Otherwise, normal initializer
        will be used. Default is ``True``.

    References
    ----------
    .. [Vaswani2017] Vaswani A, Shazeer N, Parmar N, Uszkoreit J, Jones L, Gomez AN, Kaiser Ł, Polosukhin I. Attention
        is all you need. Advances in neural information processing systems. 2017;30.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.parts.utils import rnn_weights_init
    >>> rnn = torch.nn.GRU(10, 20, 2)
    >>> rnn.apply(rnn_weights_init)
    GRU(10, 20, num_layers=2)
    """
    if isinstance(module, torch.nn.Linear):
        if xavier:
            torch.nn.init.xavier_uniform_(module.weight)
        else:
            torch.nn.init.normal_(module.weight, mean=0.0, std=std_init_range)
        if module.bias is not None:
            torch.nn.init.constant_(module.bias, 0.0)
    elif isinstance(module, torch.nn.Embedding):
        torch.nn.init.normal_(module.weight, mean=0.0, std=std_init_range)
    elif isinstance(module, torch.nn.LayerNorm):
        torch.nn.init.constant_(module.weight, 1.0)
        torch.nn.init.constant_(module.bias, 0.0)


def rss(x: torch.Tensor, dim: int = 0) -> torch.Tensor:
    """
    Compute the Root Sum of Squares (RSS).

    RSS is computed assuming that dim is the coil dimension.

    Parameters
    ----------
    x : torch.Tensor
        Tensor to apply the RSS transform to.
    dim : int, optional
        Dimension to apply the RSS transform to. Default is ``0``.

    Returns
    -------
    torch.Tensor
        Coil-combined tensor with RSS applied.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import rss
    >>> import torch
    >>> data = torch.tensor([[[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]], \
    [[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]]])
    >>> data.shape
    torch.Size([2, 2, 3, 2])
    >>> rss(data)
    tensor([[[2.8284, 2.8284],
        [5.6569, 5.6569],
        [8.4853, 8.4853]],
        [[2.8284, 2.8284],
        [5.6569, 5.6569],
        [8.4853, 8.4853]]])
    >>> rss(data).shape
    torch.Size([2, 3, 2])
    """
    return torch.sqrt((x**2).sum(dim))


def rss_complex(x: torch.Tensor, dim: int = 0) -> torch.Tensor:
    """
    Compute the Root Sum of Squares (RSS) for complex inputs.

    RSS is computed assuming that dim is the coil dimension.

    Parameters
    ----------
    x : torch.Tensor
        Tensor to apply the RSS transform to.
    dim : int, optional
        Dimension to apply the RSS transform to. Default is ``0``.

    Returns
    -------
    torch.Tensor
        Coil-combined tensor with RSS applied.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import rss_complex
    >>> import torch
    >>> data = torch.tensor([[[1+1j, 2+2j, 3+3j], [1+1j, 2+2j, 3+3j]], [[1+1j, 2+2j, 3+3j], [1+1j, 2+2j, 3+3j]]])
    >>> data.shape
    torch.Size([2, 2, 3])
    >>> rss_complex(data, dim=0)
    tensor([[1.4142, 2.8284, 4.2426],
            [1.4142, 2.8284, 4.2426]])
    >>> rss_complex(data, dim=0).shape
    torch.Size([2, 3])
    """
    return torch.sqrt(complex_abs_sq(x).sum(dim))


def save_predictions(
    predictions: Dict[str, np.ndarray], out_dir: Path, key: str = "reconstructions", file_format: str = "h5"
) -> None:
    """
    Save predictions to selected format.

    Parameters
    ----------
    predictions : Dict[str, np.ndarray]
        A dictionary mapping input filenames to corresponding predictions.
    out_dir : Path
        The output directory to save the predictions to.
    key : str, optional
        The key to save the predictions under. Default is ``reconstructions``.
    file_format : str, optional
        The format to save the predictions in. Default is ``h5``.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import save_predictions
    >>> import numpy as np
    >>> from pathlib import Path
    >>> data = {"test.h5": np.array([[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]])}
    >>> data["test.h5"].shape
    (2, 3, 2)
    >>> output_directory = Path("predictions")
    >>> save_predictions(data, output_directory, key="reconstructions", file_format="h5")
    >>> save_predictions(data, output_directory, key="segmentations", file_format="h5")
    """
    if file_format != "h5":
        raise ValueError(f"Output format {file_format} is not supported.")
    out_dir.mkdir(exist_ok=True, parents=True)
    for fname, preds in predictions.items():
        with h5py.File(out_dir / fname, "w") as hf:
            hf.create_dataset(key, data=preds)


def sense(x: torch.Tensor, sensitivity_maps: torch.Tensor, dim: int = 0) -> torch.Tensor:
    """
    Coil-combination according to the SENSitivity Encoding (SENSE) method [Pruessmann1999]_.

    References
    ----------
    .. [Pruessmann1999] Pruessmann KP, Weiger M, Scheidegger MB, Boesiger P. SENSE: Sensitivity encoding for fast MRI.
        Magn Reson Med 1999; 42:952-962.

    Parameters
    ----------
    x : torch.Tensor
        The tensor to coil-combine.
    sensitivity_maps : torch.Tensor
        The coil sensitivity maps.
    dim : int, optional
        The dimension to coil-combine along. Default is ``0``.

    Returns
    -------
    torch.Tensor
        Coil-combined tensor with SENSE applied.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import sense
    >>> import torch
    >>> data = torch.tensor([[[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]], \
    [[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]]])
    >>> data.shape
    torch.Size([2, 2, 3, 2])
    >>> coil_sensitivity_maps = torch.tensor([[[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]], \
    [[[1., 1.], [2., 2.], [3., 3.]], [[1., 1.], [2., 2.], [3., 3.]]]])
    >>> coil_sensitivity_maps.shape
    torch.Size([2, 2, 3, 2])
    >>> sense(data, coil_sensitivity_maps)
    tensor([[[2.8284, 2.8284],
        [5.6569, 5.6569],
        [8.4853, 8.4853]],
        [[2.8284, 2.8284],
        [5.6569, 5.6569],
        [8.4853, 8.4853]]])
    >>> sense(data, coil_sensitivity_maps).shape
    torch.Size([2, 3, 2])
    """
    return complex_mul(x, complex_conj(sensitivity_maps)).sum(dim)


def to_tensor(x: np.ndarray) -> torch.Tensor:
    """
    Converts a numpy array to a torch tensor. For complex arrays, the real and imaginary parts are stacked along the
    last dimension.

    Parameters
    ----------
    x : np.ndarray
        Input numpy array to be converted to torch.

    Returns
    -------
    torch.Tensor
        Torch tensor version of input.

    Examples
    --------
    >>> from atommic.collections.common.parts.utils import to_tensor
    >>> import numpy as np
    >>> data = np.array([[1+1j, 2+2j, 3+3j], [4+4j, 5+5j, 6+6j]])
    >>> data.shape
    (2, 3)
    >>> to_tensor(data)
    tensor([[[1., 1.],
            [2., 2.],
            [3., 3.]],
            [[4., 4.],
            [5., 5.],
            [6., 6.]]], dtype=torch.float64)
    >>> to_tensor(data).shape
    torch.Size([2, 3, 2])
    """
    if np.iscomplexobj(x):
        x = np.stack((x.real, x.imag), axis=-1)
    return torch.from_numpy(x)


def unnormalize(x: torch.Tensor, attrs: Dict, normalization_type: str = "max") -> torch.Tensor:
    """
    Unnormalize the input data.

    Parameters
    ----------
    x : torch.Tensor
        The input data.
    attrs : Dict
        The attributes of the input data.
    normalization_type : str
        The normalization type. Default is ``"max"``.

    Returns
    -------
    torch.Tensor
        The unnormalized data.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.parts.utils import unnormalize
    >>> data = torch.rand(1, 200, 200, 2)
    >>> attrs = {"max": 1.0, "min": 0.0}
    >>> unnormalize(data, attrs).shape
    (1, 200, 200, 2)
    """
    if normalization_type == "max":
        return x * attrs["max"]
    if normalization_type == "minmax":
        return x * (attrs["max"] - attrs["min"]) + attrs["min"]
    if normalization_type == "mean_std":
        return x * attrs["std"] + attrs["mean"]
    if normalization_type == "mean_var":
        return x * attrs["var"] + attrs["mean"]
    if normalization_type == "grayscale":
        return x / 255.0
    return x


def zero_nan_inf(x):
    """If x is nan or inf, return 0."""
    if torch.isnan(x).any() or torch.isinf(x).any():
        x = torch.tensor(0.0)
    return x



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/neural_types/comparison.py

from enum import Enum

__all__ = ["NeuralTypeComparisonResult"]


class NeuralTypeComparisonResult(Enum):
    """The result of comparing two neural type objects for compatibility. When comparing A.compare_to(B)."""

    SAME = 0
    LESS = 1  # A is B
    GREATER = 2  # B is A
    DIM_INCOMPATIBLE = 3  # Resize connector might fix incompatibility
    # A transpose and/or converting between lists and tensors will make them same
    TRANSPOSE_SAME = 4
    INCOMPATIBLE = 6  # A and B are incompatible
    # A and B are of the same type but parametrized differently
    SAME_TYPE_INCOMPATIBLE_PARAMS = 7



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/exceptions.py


class ATOMMICBaseException(Exception):
    """ATOMMIC Base Exception. All exceptions created in atommic should inherit from this class"""



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import Optional, Tuple

import torch

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import complex_conj, complex_mul


class DataConsistencyLayer(torch.nn.Module):
    """Data consistency layer for the VSNet.

    This layer is used to ensure that the output of the VSNet is the same as the input.
    """

    def __init__(self):
        """Inits :class:`DataConsistencyLayer`."""
        super().__init__()
        self.dc_weight = torch.nn.Parameter(torch.ones(1))

    def forward(self, pred_kspace: torch.Tensor, ref_kspace: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`DataConsistencyLayer`.

        Parameters
        ----------
        pred_kspace : torch.Tensor
            Predicted k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        ref_kspace : torch.Tensor
            Reference k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        """
        return ((1 - mask) * pred_kspace + mask * ref_kspace) * self.dc_weight


class WeightedAverageTerm(torch.nn.Module):
    """Weighted average term for the VSNet."""

    def __init__(self):
        """Inits :class:`WeightedAverageTerm`."""
        super().__init__()
        self.param = torch.nn.Parameter(torch.ones(1))

    def forward(self, x: torch.Tensor, Sx: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`WeightedAverageTerm`."""
        return self.param * x + (1 - self.param) * Sx


class VSNetBlock(torch.nn.Module):
    """Model block for the Variable-Splitting Network inspired by [Duan2019]_.

    References
    ----------
    .. [Duan2019] Duan, J. et al. (2019) Vs-net: Variable splitting network for accelerated parallel MRI
        reconstruction, Lecture Notes in Computer Science (including subseries Lecture Notes in Artificial
        Intelligence and Lecture Notes in Bioinformatics), 11767 LNCS, pp. 713–722. doi: 10.1007/978-3-030-32251-9_78.
    """

    def __init__(
        self,
        denoiser_block: torch.nn.ModuleList,
        data_consistency_block: torch.nn.ModuleList,
        weighted_average_block: torch.nn.ModuleList,
        num_cascades: int = 8,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
        coil_dim: int = 1,
    ):
        """Inits :class:`VSNetBlock`.

        Parameters
        ----------
        denoiser_block : torch.nn.ModuleList
            Model to apply denoising.
        data_consistency_block : torch.nn.ModuleList
            Model to apply data consistency.
        weighted_average_block : torch.nn.ModuleList
            Model to apply weighted average.
        num_cascades : int, optional
            Number of cascades. Default is ``8``.
        fft_centered : bool, optional
            Whether to center the fft. Default is ``False``.
        fft_normalization : str, optional
            The normalization of the fft. Default is ``"backward"``.
        spatial_dims : tuple, optional
            The spatial dimensions of the data. Default is ``None``.
        coil_dim : int, optional
            The dimension of the coil. Default is ``1``.
        """
        super().__init__()

        self.denoiser_block = denoiser_block
        self.data_consistency_block = data_consistency_block
        self.weighted_average_block = weighted_average_block
        self.num_cascades = num_cascades
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim

    def sens_expand(self, x: torch.Tensor, sens_maps: torch.Tensor) -> torch.Tensor:
        """Combines the sensitivity maps with coil-combined data to get multicoil data.

        Parameters
        ----------
        x : torch.Tensor
            Input data.
        sens_maps : torch.Tensor
            Coil Sensitivity maps.

        Returns
        -------
        torch.Tensor
            Expanded multicoil data.
        """
        return fft2(
            complex_mul(x, sens_maps),
            centered=self.fft_centered,
            normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
        )

    def sens_reduce(self, x: torch.Tensor, sens_maps: torch.Tensor) -> torch.Tensor:
        """Combines the sensitivity maps with multicoil data to get coil-combined data.

        Parameters
        ----------
        x : torch.Tensor
            Input data.
        sens_maps : torch.Tensor
            Coil Sensitivity maps.

        Returns
        -------
        torch.Tensor
            SENSE coil-combined reconstruction.
        """
        x = ifft2(x, centered=self.fft_centered, normalization=self.fft_normalization, spatial_dims=self.spatial_dims)
        return complex_mul(x, complex_conj(sens_maps)).sum(dim=self.coil_dim)

    def forward(self, kspace: torch.Tensor, sensitivity_maps: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`VSNetBlock`.

        Parameters
        ----------
        kspace : torch.Tensor
            Reference k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]

        Returns
        -------
        torch.Tensor
            Reconstructed image. Shape [batch_size, n_x, n_y, 2]
        """
        for idx in range(self.num_cascades):
            pred = self.sens_reduce(kspace, sensitivity_maps)
            pred = self.denoiser_block[idx](pred.permute(0, 3, 1, 2)).permute(0, 2, 3, 1)
            pred = self.sens_expand(pred, sensitivity_maps)
            sx = self.data_consistency_block[idx](pred, kspace, mask)
            sx = self.sens_reduce(sx, sensitivity_maps)
            kspace = self.weighted_average_block[idx](kspace + pred, sx)
        return kspace



# coding=utf-8
__author__ = "Dimitris Karkalousos"


import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.parts.utils import check_stacked_complex, coil_combination_method
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.didn_base.didn_block import DIDN
from atommic.collections.reconstruction.nn.sigmanet_base.dc_layers import (
    DataGDLayer,
    DataIDLayer,
    DataProxCGLayer,
    DataVSLayer,
)
from atommic.collections.reconstruction.nn.sigmanet_base.sensitivity_net import SensitivityNetwork
from atommic.collections.reconstruction.nn.unet_base.unet_block import NormUnet
from atommic.core.classes.common import typecheck

__all__ = ["DUNet"]


class DUNet(BaseMRIReconstructionModel):
    """Implementation of the Down-Up NET, inspired by [Hammernik2021]_.

    References
    ----------
    .. [Hammernik2021] Hammernik, K, Schlemper, J, Qin, C, et al. Systematic valuation of iterative deep neural
        networks for fast parallel MRI reconstruction with sensitivity-weighted coil combination. Magn Reson Med.
        2021; 86: 1859– 1872. https://doi.org/10.1002/mrm.28827

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`DUNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        reg_model_architecture = cfg_dict.get("reg_model_architecture")
        if reg_model_architecture == "DIDN":
            reg_model = DIDN(
                in_channels=cfg_dict.get("in_channels", 2),
                out_channels=cfg_dict.get("out_channels", 2),
                hidden_channels=cfg_dict.get("didn_hidden_channels"),
                num_dubs=cfg_dict.get("didn_num_dubs"),
                num_convs_recon=cfg_dict.get("didn_num_convs_recon"),
            )
        elif reg_model_architecture in ["UNET", "NORMUNET"]:
            reg_model = NormUnet(
                cfg_dict.get("unet_num_filters"),
                cfg_dict.get("unet_num_pool_layers"),
                in_chans=cfg_dict.get("in_channels", 2),
                out_chans=cfg_dict.get("out_channels", 2),
                drop_prob=cfg_dict.get("unet_dropout_probability"),
                padding_size=cfg_dict.get("unet_padding_size"),
                normalize=cfg_dict.get("unet_normalize"),
            )
        else:
            raise NotImplementedError(
                "DUNET is currently implemented for reg_model_architecture == 'DIDN' or 'UNet'."
                f"Got reg_model_architecture == {reg_model_architecture}."
            )

        data_consistency_term = cfg_dict.get("data_consistency_term")

        if data_consistency_term == "GD":
            dc_layer = DataGDLayer(
                lambda_init=cfg_dict.get("data_consistency_lambda_init"),
                fft_centered=self.fft_centered,
                fft_normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )
        elif data_consistency_term == "PROX":
            dc_layer = DataProxCGLayer(
                lambda_init=cfg_dict.get("data_consistency_lambda_init"),
                iterations=cfg_dict.get("data_consistency_iterations", 10),
                fft_centered=self.fft_centered,
                fft_normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )
        elif data_consistency_term == "VS":
            dc_layer = DataVSLayer(
                alpha_init=cfg_dict.get("data_consistency_alpha_init", 1.0),
                beta_init=cfg_dict.get("data_consistency_beta_init", 1.0),
                fft_centered=self.fft_centered,
                fft_normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )
        else:
            dc_layer = DataIDLayer()

        self.reconstruction_module = SensitivityNetwork(
            cfg_dict.get("num_iter"),
            reg_model,
            dc_layer,
            shared_params=cfg_dict.get("shared_params"),
            save_space=False,
            reset_cache=False,
        )

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`DUNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        return check_stacked_complex(
            coil_combination_method(
                self.reconstruction_module(initial_prediction, y, sensitivity_maps, mask),
                sensitivity_maps,
                self.coil_combination_method,
                self.coil_dim,
            )
        )



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/optim/lr_scheduler.py

import copy
import dataclasses
import inspect
import math
import warnings
from functools import partial
from typing import Any, Dict, Optional, Union

import hydra
import torch.optim.lr_scheduler as pt_scheduler
from omegaconf import DictConfig, OmegaConf
from torch import optim
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data import dataloader

from atommic.core.conf.schedulers import SchedulerParams, get_scheduler_config, register_scheduler_params
from atommic.utils import logging
from atommic.utils.model_utils import maybe_update_config_version


class WarmupPolicy(_LRScheduler):
    """Adds warmup kwargs and warmup logic to lr policy. All arguments should be passed as kwargs for clarity.

    Returns
    -------
    lr : float
        Learning rate for current step.
    """

    def __init__(self, optimizer, *, warmup_steps=None, warmup_ratio=None, max_steps=None, min_lr=0.0, last_epoch=-1):
        """Inits :class:`WarmupPolicy`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer.
        warmup_steps : int
            Number of training steps in warmup stage. Default is ``None``.
        warmup_ratio : float
            Ratio of warmup steps to total steps. Default is ``None``.
        max_steps : int
            Total number of steps while training or `None` for infinite training. Default is ``None``.
        min_lr : float
            Minimum learning rate. Default is ``0``.
        last_epoch : int
            Last epoch. Default is ``-1``.
        """
        if warmup_steps is not None and warmup_ratio is not None:
            raise AssertionError("Either use particular number of step or ratio")
        if warmup_ratio is not None and max_steps is None:
            raise AssertionError("If there is a ratio, there should be a total steps")

        # It is necessary to assign all attributes *before* __init__,
        # as class is wrapped by an inner class.
        self.max_steps = max_steps
        if warmup_steps is not None:
            self.warmup_steps = warmup_steps
        elif warmup_ratio is not None:
            self.warmup_steps = int(warmup_ratio * max_steps)
        else:
            self.warmup_steps = 0

        self.min_lr = min_lr
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        """Get learning rate at current step."""
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, please use `get_last_lr()`.", UserWarning
            )

        step = self.last_epoch

        if step <= self.warmup_steps and self.warmup_steps > 0:
            return self._get_warmup_lr(step)

        if (self.max_steps is not None) and (step > self.max_steps):
            return [self.min_lr for _ in self.base_lrs]

        return self._get_lr(step)

    def _get_warmup_lr(self, step):
        """Linear warmup"""
        lr_val = (step + 1) / (self.warmup_steps + 1)
        return [initial_lr * lr_val for initial_lr in self.base_lrs]

    def _get_lr(self, step):  # pylint: disable=unused-argument
        """Simple const lr policy"""
        return self.base_lrs


class SquareRootConstantPolicy(_LRScheduler):
    """Adds warmup kwargs and warmup logic to lr policy. All arguments should be passed as kwargs for clarity."""

    def __init__(
        self, optimizer, *, constant_steps=None, constant_ratio=None, max_steps=None, min_lr=0.0, last_epoch=-1
    ):
        """Inits :class:`SquareRootConstantPolicy`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer.
        constant_steps : int
            Number of training steps in constant stage. Default is ``None``.
        constant_ratio : float
            Ratio of constant steps to total steps. Default is ``None``.
        max_steps : int
            Total number of steps while training or `None` for infinite training. Default is ``None``.
        min_lr : float
            Minimum learning rate. Default is ``0``.
        last_epoch : int
            Last epoch. Default is ``-1``.
        """
        if constant_steps is not None and constant_ratio is not None:
            raise AssertionError("Either use particular number of step or ratio")

        if constant_ratio is not None and max_steps is None:
            raise AssertionError("If there is a ratio, there should be a total steps")

        # It is necessary to assign all attributes *before* __init__, as class is wrapped by an inner class.
        self.max_steps = max_steps
        if constant_steps is not None:
            self.constant_steps = constant_steps
        elif constant_ratio is not None:
            self.constant_steps = int(constant_ratio * max_steps)
        else:
            self.constant_steps = 0

        self.constant_lr = 1 / (constant_steps**0.5)
        self.min_lr = min_lr
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        """Get learning rate at current step."""
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, please use `get_last_lr()`.", UserWarning
            )

        step = self.last_epoch

        if step <= self.constant_steps:
            return [self.constant_lr for _ in self.base_lrs]

        if step > self.max_steps:
            return [self.min_lr for _ in self.base_lrs]

        return self._get_lr(step)

    def _get_lr(self, step):  # pylint: disable=unused-argument
        """Simple const lr policy"""
        return self.base_lrs


class WarmupHoldPolicy(WarmupPolicy):
    """Variant of WarmupPolicy which maintains high learning rate for a defined number of steps. All arguments should
    be passed as kwargs for clarity,

    Results
    -------
    lr : float
        Learning rate is linearly increased from 0 to 1 over warmup steps, then linearly decreased from 1 to 0 over
        hold steps.
    """

    def __init__(
        self,
        optimizer,
        *,
        warmup_steps=None,
        warmup_ratio=None,
        hold_steps=None,
        hold_ratio=None,
        max_steps=None,
        min_lr=0.0,
        last_epoch=-1,
    ):
        """Inits :class:`WarmupHoldPolicy`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer.
        warmup_steps : int
            Number of training steps in warmup stage. Default is ``None``.
        warmup_ratio : float
            Ratio of warmup steps to total steps. Default is ``None``.
        hold_steps : int
            Number of training steps to hold the learning rate after warm up. Default is ``None``.
        hold_ratio : float
            Ratio of hold steps to total steps. Default is ``None``.
        max_steps : int
            Total number of steps while training or `None` for infinite training. Default is ``None``.
        min_lr : float
            Minimum learning rate. Default is ``0``.
        last_epoch : int
            Last epoch. Default is ``-1``.
        """
        if hold_steps is not None and hold_ratio is not None:
            raise AssertionError("Either use particular number of step or ratio")
        if hold_ratio is not None and max_steps is None:
            raise AssertionError("If there is a ratio, there should be a total steps")

        self.min_lr = min_lr
        self._last_warmup_lr = 0.0

        # Necessary to duplicate as class attributes are hidden in inner class
        self.max_steps = max_steps
        if warmup_steps is not None:
            self.warmup_steps = warmup_steps
        elif warmup_ratio is not None:
            self.warmup_steps = int(warmup_ratio * max_steps)
        else:
            self.warmup_steps = 0

        if hold_steps is not None:
            self.hold_steps = hold_steps + self.warmup_steps
        elif hold_ratio is not None:
            self.hold_steps = int(hold_ratio * max_steps) + self.warmup_steps
        else:
            self.hold_steps = 0

        super().__init__(
            optimizer,
            warmup_steps=warmup_steps,
            warmup_ratio=warmup_ratio,
            max_steps=max_steps,
            last_epoch=last_epoch,
            min_lr=min_lr,
        )

    def get_lr(self):
        """Get learning rate at current step."""
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, please use `get_last_lr()`.", UserWarning
            )

        step = self.last_epoch

        # Warmup phase
        if 0 < self.warmup_steps >= step:
            return self._get_warmup_lr(step)

        # Hold phase
        if self.hold_steps < step >= self.warmup_steps:
            return self.base_lrs

        if step > self.max_steps:
            return [self.min_lr for _ in self.base_lrs]

        return self._get_lr(step)


class WarmupAnnealHoldPolicy(_LRScheduler):
    """Adds warmup kwargs and warmup logic to lr policy. All arguments should be passed as kwargs for clarity."""

    def __init__(
        self,
        optimizer,
        *,
        warmup_steps=None,
        warmup_ratio=None,
        constant_steps=None,
        constant_ratio=None,
        max_steps=None,
        min_lr=0.0,
        last_epoch=-1,
    ):
        """Inits :class:`WarmupAnnealHoldPolicy`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer.
        warmup_steps : int
            Number of training steps in warmup stage. Default is ``None``.
        warmup_ratio : float
            Ratio of warmup steps to total steps. Default is ``None``.
        constant_steps : int
            Number of training steps in constant stage. Default is ``None``.
        constant_ratio : float
            Ratio of constant steps to total steps. Default is ``None``.
        max_steps : int
            Total number of steps while training or `None` for infinite training. Default is ``None``.
        min_lr : float
            Minimum learning rate. Default is ``0``.
        last_epoch : int
            Last epoch. Default is ``-1``.
        """
        if warmup_steps is not None and warmup_ratio is not None:
            raise AssertionError("Either use particular number of step or ratio")
        if constant_steps is not None and constant_ratio is not None:
            raise AssertionError("Either use constant_steps or constant_ratio")
        if warmup_ratio is not None and max_steps is None:
            raise AssertionError("If there is a ratio, there should be a total steps")

        # It is necessary to assign all attributes *before* __init__, as class is wrapped by an inner class.
        self.max_steps = max_steps

        if warmup_steps is not None:
            self.warmup_steps = warmup_steps
        elif warmup_ratio is not None:
            self.warmup_steps = int(warmup_ratio * max_steps)
        else:
            self.warmup_steps = 0

        if constant_steps is not None:
            self.constant_steps = constant_steps
        elif constant_ratio is not None:
            self.constant_steps = int(constant_ratio * max_steps)
        else:
            self.constant_steps = 0

        self.decay_steps = max_steps - (self.constant_steps + self.warmup_steps)

        self.min_lr = min_lr
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        """Get learning rate at current step."""
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, please use `get_last_lr()`.", UserWarning
            )

        step = self.last_epoch

        # Warmup steps
        if 0 < self.warmup_steps >= step:
            return self._get_warmup_lr(step)

        # Constant steps after warmup and decay
        if self.constant_steps > 0 and (self.warmup_steps + self.decay_steps) < step <= self.max_steps:
            return self._get_constant_lr(step)

        # Min lr after max steps of updates
        if step > self.max_steps:
            return [self.min_lr for _ in self.base_lrs]

        return self._get_lr(step)

    def _get_warmup_lr(self, step):
        """Get learning rate at warmup stage."""
        lr_val = (step + 1) / (self.warmup_steps + 1)
        return [initial_lr * lr_val for initial_lr in self.base_lrs]

    def _get_constant_lr(self, step):  # pylint: disable=unused-argument
        """Get learning rate at constant stage."""
        return [self.min_lr for _ in self.base_lrs]

    def _get_lr(self, step):  # pylint: disable=unused-argument
        """Simple const lr policy"""
        return self.base_lrs


def _sqrt_annealing(initial_lr, step, max_steps, min_lr):
    """Anneal learning rate by sqrt."""
    mult = ((max_steps - step) / max_steps) ** 0.5
    out_lr = initial_lr * mult
    out_lr = max(out_lr, min_lr)
    return out_lr


def _square_annealing(initial_lr, step, max_steps, min_lr):
    """Anneal learning rate by square."""
    mult = ((max_steps - step) / max_steps) ** 2
    out_lr = initial_lr * mult
    out_lr = max(out_lr, min_lr)
    return out_lr


def _cosine_annealing(initial_lr, step, max_steps, min_lr):
    """Anneal learning rate by cosine."""
    mult = 0.5 * (1 + math.cos(math.pi * step / max_steps))
    return (initial_lr - min_lr) * mult + min_lr


def _linear_warmup_with_cosine_annealing(max_lr, warmup_steps, step, decay_steps, min_lr):
    """Anneal learning rate by linear warmup and cosine annealing."""
    if max_lr <= min_lr:
        raise AssertionError
    # Use linear warmup for the initial part.
    if warmup_steps > 0 and step <= warmup_steps:
        return max_lr * float(step) / float(warmup_steps)

    # For any steps larger than `decay_steps`, use `min_lr`.
    if step > warmup_steps + decay_steps:
        return min_lr

    # If we are done with the warmup period, use the decay style.
    num_steps_ = step - warmup_steps
    decay_steps_ = decay_steps
    decay_ratio = float(num_steps_) / float(decay_steps_)
    if decay_ratio < 0.0:
        raise AssertionError
    if decay_ratio > 1.0:
        raise AssertionError
    delta_lr = max_lr - min_lr

    coeff = 0.5 * (math.cos(math.pi * decay_ratio) + 1.0)

    return min_lr + coeff * delta_lr


def _poly_decay(initial_lr, step, decay_steps, power, min_lr, cycle):
    """Polynomial decay of learning rate."""
    if cycle:
        multiplier = 1.0 if step == 0 else math.ceil(step / decay_steps)
        decay_steps = decay_steps * multiplier
    else:
        step = min(step, decay_steps)
    p = step / decay_steps
    lr = (initial_lr - min_lr) * math.pow(1.0 - p, power)
    lr += min_lr
    return lr


def _noam_hold_annealing(initial_lr, step, warmup_steps, hold_steps, decay_rate, min_lr):
    """Anneal learning rate by noam hold."""
    # hold_steps = total number of steps to hold the LR, not the warmup + hold steps.
    T_warmup_decay = max(1, warmup_steps**decay_rate)
    T_hold_decay = max(1, (step - hold_steps) ** decay_rate)
    lr = (initial_lr * T_warmup_decay) / T_hold_decay
    return max(lr, min_lr)


class SquareAnnealing(WarmupPolicy):
    """Anneal learning rate by square."""

    def __init__(self, optimizer, *, max_steps, min_lr=1e-5, last_epoch=-1, **kwargs):
        """Inits :class:`SquareAnnealing`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer.
        max_steps : int
            Total number of steps while training or `None` for infinite training.
        min_lr : float
            Minimum learning rate. Default is ``1e-5``.
        last_epoch : int
            Last epoch. Default is ``-1``.
        """
        super().__init__(optimizer=optimizer, max_steps=max_steps, last_epoch=last_epoch, min_lr=min_lr, **kwargs)

    def _get_lr(self, step):
        """Get learning rate at current step."""
        return [
            _square_annealing(
                initial_lr=initial_lr,
                step=step - self.warmup_steps,
                max_steps=self.max_steps - self.warmup_steps,
                min_lr=self.min_lr,
            )
            for initial_lr in self.base_lrs
        ]


class SquareRootAnnealing(WarmupPolicy):
    """Anneal learning rate by square root."""

    def __init__(self, optimizer, *, max_steps, min_lr=0, last_epoch=-1, **kwargs):
        """Inits :class:`SquareRootAnnealing`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer.
        max_steps : int
            Total number of steps while training or `None` for infinite training.
        min_lr : float
            Minimum learning rate. Default is ``0``.
        last_epoch : int
            Last epoch. Default is ``-1``.
        """
        super().__init__(optimizer=optimizer, max_steps=max_steps, last_epoch=last_epoch, min_lr=min_lr, **kwargs)

    def _get_lr(self, step):
        """Get learning rate at current step."""
        return [
            _sqrt_annealing(
                initial_lr=initial_lr,
                step=step,
                max_steps=self.max_steps,
                min_lr=self.min_lr,
            )
            for initial_lr in self.base_lrs
        ]


class CosineAnnealing(WarmupAnnealHoldPolicy):
    """Anneal learning rate by cosine."""

    def __init__(self, optimizer, *, max_steps, min_lr=0, last_epoch=-1, **kwargs):
        """Inits :class:`CosineAnnealing`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer.
        max_steps : int
            Total number of steps while training or `None` for infinite training.
        min_lr : float
            Minimum learning rate. Default is ``0``.
        last_epoch : int
            Last epoch. Default is ``-1``.
        """
        super().__init__(optimizer=optimizer, max_steps=max_steps, last_epoch=last_epoch, min_lr=min_lr, **kwargs)

    def _get_lr(self, step):
        """Get learning rate at current step."""
        for initial_lr in self.base_lrs:
            if initial_lr < self.min_lr:
                raise ValueError(
                    f"{self} received an initial learning rate that was lower than the minimum learning rate."
                )

        return (
            [
                _cosine_annealing(
                    initial_lr=initial_lr,
                    step=step - self.warmup_steps,
                    max_steps=self.max_steps - self.warmup_steps,
                    min_lr=self.min_lr,
                )
                for initial_lr in self.base_lrs
            ]
            if self.constant_steps is None or self.constant_steps == 0
            else self._get_linear_warmup_with_cosine_annealing_lr(step)
        )

    def _get_warmup_lr(self, step):
        """Get the warmup learning rate for the given step."""
        if self.constant_steps is None or self.constant_steps == 0:
            return super()._get_warmup_lr(step)

        # Use linear warmup for the initial part.
        return self._get_linear_warmup_with_cosine_annealing_lr(step)

    def _get_constant_lr(self, step):
        """Only called when constant_steps is not None and not 0."""
        return self._get_linear_warmup_with_cosine_annealing_lr(step)

    def _get_linear_warmup_with_cosine_annealing_lr(self, step):
        """Cosine Schedule, slightly different warmup schedule + constant LR at the end."""
        return [
            _linear_warmup_with_cosine_annealing(
                max_lr=self.base_lrs[0],
                warmup_steps=self.warmup_steps,
                step=step,
                decay_steps=self.decay_steps,
                min_lr=self.min_lr,
            )
            for _ in self.base_lrs
        ]


class NoamAnnealing(_LRScheduler):
    """Noam learning rate annealing."""

    def __init__(
        self, optimizer, *, d_model, warmup_steps=None, warmup_ratio=None, max_steps=None, min_lr=0.0, last_epoch=-1
    ):
        """Inits :class:`NoamAnnealing`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer.
        d_model : int
            Model dimensionality.
        warmup_steps : int
            Number of training steps in warmup stage. Default is ``None``.
        warmup_ratio : float
            Ratio of warmup steps to total steps. Default is ``None``.
        max_steps : int
            Total number of steps while training or `None` for infinite training.
        min_lr : float
            Minimum learning rate. Default is ``0``.
        last_epoch : int
            Last epoch. Default is ``-1``.
        """
        self._normalize = d_model ** (-0.5)
        if warmup_steps is not None and warmup_ratio is not None:
            raise AssertionError("Either use particular number of step or ratio")
        if warmup_ratio is not None and max_steps is None:
            raise AssertionError("If there is a ratio, there should be a total steps")

        # It is necessary to assign all attributes *before* __init__,
        # as class is wrapped by an inner class.
        self.max_steps = max_steps
        if warmup_steps is not None:
            self.warmup_steps = warmup_steps
        elif warmup_ratio is not None:
            self.warmup_steps = int(warmup_ratio * max_steps)
        else:
            self.warmup_steps = 0

        self.min_lr = min_lr
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        """Get learning rate at current step."""
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, please use `get_last_lr()`.", UserWarning
            )

        step = max(1, self.last_epoch)

        if step > self.max_steps:
            return [self.min_lr for _ in self.base_lrs]

        for initial_lr in self.base_lrs:
            if initial_lr < self.min_lr:
                raise ValueError(
                    f"{self} received an initial learning rate that was lower than the minimum learning rate."
                )

        return [self._noam_annealing(initial_lr=initial_lr, step=step) for initial_lr in self.base_lrs]

    def _noam_annealing(self, initial_lr, step):
        """Noam learning rate annealing."""
        mult = (
            self._normalize * min(step ** (-0.5), step * (self.warmup_steps ** (-1.5)))
            if self.warmup_steps > 0
            else self._normalize * step ** (-0.5)
        )
        out_lr = initial_lr * mult
        if step > self.warmup_steps:
            out_lr = max(out_lr, self.min_lr)
        return out_lr


class NoamHoldAnnealing(WarmupHoldPolicy):
    """Implementation of the Noam Hold Annealing policy from the SqueezeFormer paper.

    Unlike NoamAnnealing, the peak learning rate can be explicitly set for this scheduler.
    The schedule first performs linear warmup, then holds the peak LR, then decays with some schedule for
    the remainder of the steps. Therefore, the min-lr is still dependent on the hyperparameters selected.

    It's schedule is determined by three factors-

    Warmup Steps: Initial stage, where linear warmup occurs uptil the peak LR is reached. Unlike NoamAnnealing,
        the peak LR is explicitly stated here instead of a scaling factor.

    Hold Steps: Intermediate stage, where the peak LR is maintained for some number of steps. In this region,
        the high peak LR allows the model to converge faster if training is stable. However the high LR
        may also cause instability during training. Should usually be a significant fraction of training
        steps (around 30-40% of the entire training steps).

    Decay Steps: Final stage, where the LR rapidly decays with some scaling rate (set by decay rate).
        To attain Noam decay, use 0.5, for Squeezeformer recommended decay, use 1.0. The fast decay after
        prolonged high LR during hold phase allows for rapid convergence.

    References
    ----------
        [1]
        [Squeezeformer: An Efficient Transformer for Automatic Speech Recognition](https://arxiv.org/abs/2206.00888)
    """

    def __init__(self, optimizer, *, max_steps, decay_rate=0.5, min_lr=0.0, last_epoch=-1, **kwargs):
        """Inits :class:`NoamHoldAnnealing`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer to use for the scheduler.
        max_steps : int
            Total number of training steps.
        decay_rate : float
            Decay rate for the final stage of the schedule. Should be between 0 and 1. Default is ``0.5``.
        min_lr : float
            Minimum learning rate to use for the schedule. Should be between 0 and 1. Default is ``0.0``.
        last_epoch : int
            Last epoch to start the schedule from. Should be between 0 and max_steps. Default is ``-1``.
        """
        self.decay_rate = decay_rate
        super().__init__(optimizer=optimizer, max_steps=max_steps, last_epoch=last_epoch, min_lr=min_lr, **kwargs)

    def _get_lr(self, step):
        """Get the learning rate for the given step."""
        if self.warmup_steps is None or self.warmup_steps == 0:
            raise ValueError("Noam scheduler cannot be used without warmup steps")

        if self.hold_steps > 0:
            hold_steps = self.hold_steps - self.warmup_steps
        else:
            hold_steps = 0

        return [
            _noam_hold_annealing(
                initial_lr,
                step=step,
                warmup_steps=self.warmup_steps,
                hold_steps=hold_steps,
                decay_rate=self.decay_rate,
                min_lr=self.min_lr,
            )
            for initial_lr in self.base_lrs
        ]


class WarmupAnnealing(WarmupPolicy):
    """Warmup learning rate annealing."""

    def __init__(self, optimizer, *, max_steps, last_epoch=-1, min_lr=0.0, **kwargs):
        """Inits :class:`WarmupAnnealing`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer to use for the scheduler.
        max_steps : int
            Total number of training steps.
        last_epoch : int
            Last epoch to start the schedule from. Should be between 0 and max_steps. Default is ``-1``.
        min_lr : float
            Minimum learning rate to use for the schedule. Should be between 0 and 1. Default is ``0.0``.
        """
        super().__init__(optimizer=optimizer, max_steps=max_steps, last_epoch=last_epoch, min_lr=min_lr, **kwargs)

    def _get_lr(self, step):
        """Get learning rate at current step."""
        delta_lr = self.base_lrs[0] - self.min_lr
        mult = (step - self.warmup_steps) / (self.max_steps - self.warmup_steps)
        return [self.min_lr + (1 - mult) * delta_lr for _ in self.base_lrs]


class InverseSquareRootAnnealing(WarmupPolicy):
    """Inverse square root learning rate annealing."""

    def __init__(self, optimizer, *, max_steps, last_epoch=-1, min_lr=0.0, **kwargs):
        """Inits :class:`InverseSquareRootAnnealing`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer to use for the scheduler.
        max_steps : int
            Total number of training steps.
        last_epoch : int
            Last epoch to start the schedule from. Should be between 0 and max_steps. Default is ``-1``.
        min_lr : float
            Minimum learning rate to use for the schedule. Should be between 0 and 1. Default is ``0.0``.
        """
        super().__init__(optimizer=optimizer, max_steps=max_steps, **kwargs, last_epoch=last_epoch, min_lr=min_lr)

    def _get_lr(self, step):
        """Get learning rate at current step."""
        denom = ((step + 1) / (self.warmup_steps + 1)) ** 0.5
        return [initial_lr / denom for initial_lr in self.base_lrs]


class T5InverseSquareRootAnnealing(SquareRootConstantPolicy):
    """Inverse square root learning rate annealing."""

    def __init__(self, optimizer, *, max_steps, last_epoch=-1, min_lr=0.0, **kwargs):
        """Inits :class:`T5InverseSquareRootAnnealing`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer to use for the scheduler.
        max_steps : int
            Total number of training steps.
        last_epoch : int
            Last epoch to start the schedule from. Should be between 0 and max_steps. Default is ``-1``.
        min_lr : float
            Minimum learning rate to use for the schedule. Should be between 0 and 1. Default is ``0.0``.
        """
        super().__init__(optimizer=optimizer, max_steps=max_steps, **kwargs, last_epoch=last_epoch, min_lr=min_lr)

    def _get_lr(self, step):
        """Get learning rate at current step."""
        return [1 / (step**0.5) for _ in self.base_lrs]


class PolynomialDecayAnnealing(WarmupPolicy):
    """Polynomial decay learning rate annealing."""

    def __init__(self, optimizer, *, max_steps, min_lr=0.0, power=1.0, cycle=False, last_epoch=-1, **kwargs):
        """Inits :class:`PolynomialDecayAnnealing`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer to use for the scheduler.
        max_steps : int
            Total number of training steps.
        min_lr : float
            Minimum learning rate to use for the schedule. Should be between 0 and 1. Default is ``0.0``.
        power : float
            Power of the polynomial. Default is ``1.0``.
        cycle : bool
            Whether to cycle the schedule. Default is ``False``.
        last_epoch : int
            Last epoch to start the schedule from. Should be between 0 and max_steps. Default is ``-1``.
        """
        self.power = power
        self.cycle = cycle

        super().__init__(optimizer=optimizer, max_steps=max_steps, last_epoch=last_epoch, min_lr=min_lr, **kwargs)

    def _get_lr(self, step):
        """Get learning rate at current step."""
        return [
            _poly_decay(
                initial_lr,
                step=step - self.warmup_steps,
                decay_steps=self.max_steps - self.warmup_steps,
                power=self.power,
                min_lr=self.min_lr,
                cycle=self.cycle,
            )
            for initial_lr in self.base_lrs
        ]


class PolynomialHoldDecayAnnealing(WarmupHoldPolicy):
    """Polynomial decay learning rate annealing."""

    def __init__(self, optimizer, *, max_steps, min_lr=0.0, power=1.0, cycle=False, last_epoch=-1, **kwargs):
        """Inits :class:`PolynomialHoldDecayAnnealing`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer to use for the scheduler.
        max_steps : int
            Total number of training steps.
        min_lr : float
            Minimum learning rate to use for the schedule. Should be between 0 and 1. Default is ``0.0``.
        power : float
            Power of the polynomial. Default is ``1.0``.
        cycle : bool
            Whether to cycle the schedule. Default is ``False``.
        last_epoch : int
            Last epoch to start the schedule from. Should be between 0 and max_steps. Default is ``-1``.
        """
        self.power = power
        self.cycle = cycle

        super().__init__(optimizer=optimizer, max_steps=max_steps, last_epoch=last_epoch, min_lr=min_lr, **kwargs)

    def _get_lr(self, step):
        """Get learning rate at current step."""
        return [
            _poly_decay(
                initial_lr,
                step=step - self.hold_steps,
                decay_steps=self.max_steps - max(self.warmup_steps, self.hold_steps),
                power=self.power,
                min_lr=self.min_lr,
                cycle=self.cycle,
            )
            for initial_lr in self.base_lrs
        ]


def register_scheduler(name: str, scheduler: _LRScheduler, scheduler_params: SchedulerParams):
    """Checks if the scheduler name exists in the registry, and if it doesn't, adds it. This allows custom schedulers
    to be added and called by name during instantiation.

    Parameters
    ----------
    name : str
        Name of the optimizer. Will be used as key to retrieve the optimizer.
    scheduler : _LRScheduler
        Scheduler class (inherits from _LRScheduler)
    scheduler_params : SchedulerParams
        The parameters as a dataclass of the scheduler
    """
    if name in AVAILABLE_SCHEDULERS:
        raise ValueError(f"Cannot override pre-existing schedulers. Conflicting scheduler name = {name}")

    AVAILABLE_SCHEDULERS[name] = scheduler

    sched_name = f"{scheduler.__name__}_params"
    register_scheduler_params(name=sched_name, scheduler_params=scheduler_params)


def get_scheduler(name: str, **kwargs: Optional[Dict[str, Any]]) -> _LRScheduler:
    """Convenience method to obtain an _LRScheduler class and partially instantiate it with optimizer kwargs.

    Parameters
    ----------
    name : str
        Name of the scheduler in the registry.
    kwargs : Dict[str, Any]
        Optional kwargs of the scheduler used during instantiation.

    Returns
    -------
    _LRScheduler
        A partially instantiated _LRScheduler.
    """
    if name not in AVAILABLE_SCHEDULERS:
        raise ValueError(
            f"Cannot resolve scheduler{name}'. Available optimizers are : " f"{AVAILABLE_SCHEDULERS.keys()}"
        )

    scheduler_cls = AVAILABLE_SCHEDULERS[name]
    # Pop 'max_steps' if it's not required by the scheduler
    if "max_steps" in kwargs and "max_steps" not in inspect.signature(scheduler_cls).parameters:
        kwargs.pop("max_steps")
    scheduler = partial(scheduler_cls, **kwargs)
    return scheduler


def prepare_lr_scheduler(  # noqa: MC0001
    optimizer: optim.Optimizer,
    scheduler_config: Union[Dict[str, Any], DictConfig, None],
    train_dataloader: Optional[dataloader.DataLoader] = None,
) -> Optional[Dict[str, Any]]:
    """Constructs an LR Scheduler (optionally) for a given optimizer, based on a config with the following schema.

    Parameters
    ----------
    optimizer: The optimizer to use for the scheduler.
        name: <name of optimizer>
        lr: <maximal learning rate>
        args:
            name: auto  # special keyword, resolves to correct optimizer config for given optimizer name
            # cls: atommic.core.config.optimizers.NovogradParams  # explicit instantiation by class path
            params:  # optional override parameters for the optimizer config
                betas: [0.8, 0.5]
                weight_decay: 0.001
    scheduler_config: The scheduler config.
        name: <name of scheduler>
        iters_per_batch: null # computed at runtime; mandatory to have
        max_steps: null # computed at runtime or explicitly set here; mandatory to have
        # pytorch lightning args <mandatory>
        monitor: val_loss
        reduce_on_plateau: false
        # <scheduler config override>
        args:
            name: auto  # special keyword, resolves to correct optimizer config for given optimizer name
            # cls: atommic.core.config.schedulers.CosineAnnealingParams  # explicit instantiation by class path
            params:  # optional override parameters for the optimizer config
                warmup_steps: null
                warmup_ratio: null
                min_lr: 0.0
                last_epoch: -1

    train_dataloader: Optional requirement, must be passed if "iters_per_batch" is defined instead of "max_steps". \
    Used to compute effective "max_steps".

    Returns
    -------
    A dictionary containing the LR Scheduler implementation if the config was successfully parsed along with other \
    parameters required by Pytorch Lightning, otherwise None.
    """
    if scheduler_config is not None:
        scheduler_config = maybe_update_config_version(scheduler_config)

    # Build nested dictionary for convenience out of structured objects
    if isinstance(scheduler_config, DictConfig):
        scheduler_config = OmegaConf.to_container(scheduler_config, resolve=True)

    elif dataclasses.is_dataclass(scheduler_config):
        # Recursively transform data classes to basic dictionaries
        scheduler_config = OmegaConf.create(scheduler_config)
        scheduler_config = OmegaConf.to_container(scheduler_config, resolve=True)

    # Test to see if config follows above schema
    interval = "step"
    if scheduler_config is not None:
        if "args" in scheduler_config:
            scheduler_args = scheduler_config.pop("args")
        else:
            scheduler_args = copy.deepcopy(scheduler_config)

            # Remove extra parameters from scheduler_args nest
            # Assume all other parameters are to be passed into scheduler constructor
            scheduler_args.pop("name", None)
            scheduler_args.pop("t_max_epochs", None)
            scheduler_args.pop("t_accumulate_grad_batches", None)
            scheduler_args.pop("t_limit_train_batches", None)
            scheduler_args.pop("t_num_workers", None)
            scheduler_args.pop("monitor", None)
            scheduler_args.pop("reduce_on_plateau", None)

        if "name" in scheduler_config and scheduler_config["name"] in EPOCH_SCHEDULERS:
            interval = "epoch"
    else:
        # Return gracefully in case `sched` was not supplied; inform user
        logging.info("Scheduler not initialized as no `sched` config supplied to setup_optimizer()")
        return None

    # Try instantiation of scheduler params from config class path
    if "_target_" in scheduler_args:
        scheduler_args_cfg = OmegaConf.create(scheduler_args)
        scheduler_conf = hydra.utils.instantiate(scheduler_args_cfg)
        scheduler_args = vars(scheduler_conf)

        # Get name of the scheduler
        scheduler_name = scheduler_conf.__class__.__name__

        if "Params" in scheduler_name:
            scheduler_name = scheduler_name.replace("Params", "")

    else:
        # Class path instantiation failed; try resolving "name" component

        # Get name of the scheduler
        if "name" in scheduler_config:
            scheduler_name = scheduler_config["name"]
        else:
            logging.warning(
                "Could not resolve classpath for Scheduler Config, and `name` "
                "was not provided either. \n"
                "Scheduler cannot be instantiated !"
            )
            return None

        # If class path was not provided, perhaps `name` is provided for resolution
        if "name" in scheduler_args:
            # If `auto` is passed as name for resolution of optimizer name,
            # then lookup optimizer name and resolve its parameter config
            if scheduler_args["name"] == "auto":
                scheduler_params_name = f"{scheduler_name}Params"
            else:
                scheduler_params_name = scheduler_args["name"]

            # Get override arguments provided in the config yaml file / Dict Config
            scheduler_params_override = scheduler_args.get("params", {})

            # If params is itself a dict config object provided explicitly in Dict Config
            # Resolve to dictionary for convenience
            if isinstance(scheduler_params_override, DictConfig):
                scheduler_params_override = OmegaConf.to_container(scheduler_params_override, resolve=True)

            # Get and instantiate the Config dataclass for this scheduler
            scheduler_params_cls = get_scheduler_config(scheduler_params_name, **scheduler_params_override)
            scheduler_params = scheduler_params_cls  # instantiate the parameters object
            # extract just the dictionary from the Config object
            scheduler_args = vars(scheduler_params)

    # Extract value to monitor in losses, if provided.
    if "monitor" in scheduler_config:
        monitor = scheduler_config.get("monitor")
    else:
        # Default to train loss
        monitor = "loss"

    # Store exact max_steps if it is provided
    if "max_steps" in scheduler_config and scheduler_config["max_steps"] is not None:
        max_steps = scheduler_config["max_steps"]
    elif "t_max_epochs" in scheduler_config:
        # Compute effective max_steps if t_max_epochs is provided
        if train_dataloader is None:
            logging.warning(
                "As `t_max_epochs` is provided/computed, it is required to pass the train dataloader in order\n"
                "to compute effective maximum number of steps.\n"
                "Scheduler will not be instantiated !"
            )
            return None

        # Raise exception if neither `max_steps` nor `t_max_epochs` is provided
        if scheduler_config.get("t_max_epochs", None) is None:
            logging.warning(
                "`t_max_epochs` cannot be None when `max_steps` is not not provided.\n"
                "This can occur when `train dataloader` is not available to correctly "
                "prepare the scheduler.\n"
                "Scheduler will not be instantiated !"
            )
            return None

        # Get iters_per_batch
        max_epochs = scheduler_config.get("t_max_epochs")
        accumulate_grad_batches = scheduler_config.get("t_accumulate_grad_batches")
        limit_train_batches = scheduler_config.get("t_limit_train_batches")
        num_workers = scheduler_config.get("t_num_workers")

        # Compute effective num max_steps
        num_samples = len(train_dataloader.dataset)

        # we may need to override ModelPT setup_optimization
        if train_dataloader.batch_size is not None:
            batch_size = train_dataloader.batch_size
        elif hasattr(train_dataloader, "batch_sampler") and train_dataloader.batch_sampler is not None:
            if train_dataloader.batch_sampler.micro_batch_size is not None:
                batch_size = train_dataloader.batch_sampler.micro_batch_size
            else:
                raise ValueError(f"Could not find batch_size from batch_sampler: {train_dataloader.batch_sampler}")
        else:
            raise ValueError(f"Could not find batch_size from train_dataloader: {train_dataloader}")
        drop_last = train_dataloader.drop_last

        max_steps = compute_max_steps(
            max_epochs=max_epochs,
            accumulate_grad_batches=accumulate_grad_batches,
            limit_train_batches=limit_train_batches,
            num_workers=num_workers,
            num_samples=num_samples,
            batch_size=batch_size,
            drop_last=drop_last,
        )
    else:
        logging.warning(
            "Neither `max_steps` nor `iters_per_batch` were provided to `optim.sched`, "
            "cannot compute effective `max_steps` !\n"
            "Scheduler will not be instantiated !"
        )
        return None

    # Inject max_steps (effective or provided) into the scheduler config
    scheduler_args["max_steps"] = max_steps

    # Get the scheduler class from the config
    scheduler_cls = get_scheduler(scheduler_name, **scheduler_args)

    # Pop 'max_steps' if it's not required by the scheduler
    if "max_steps" not in inspect.signature(scheduler_cls).parameters:
        scheduler_args.pop("max_steps")

    # Instantiate the LR schedule
    schedule = scheduler_cls(optimizer, **scheduler_args)

    logging.info(
        'Scheduler "%s" \nwill be used during training (effective maximum steps = %d) - \nParameters : \n(%s)',
        str(schedule),
        max_steps,
        OmegaConf.to_yaml(OmegaConf.create(scheduler_args)),
    )

    # Wrap the schedule in PTL arguments to perform stepwise computation. Rather than epoch level computation.
    reduce_lr_on_plateau = bool(isinstance(schedule, optim.lr_scheduler.ReduceLROnPlateau))

    schedule_dict = {
        "scheduler": schedule,
        "interval": interval,
        "frequency": 1,
        "monitor": monitor,
        "reduce_on_plateau": reduce_lr_on_plateau,
    }
    return schedule_dict


def compute_max_steps(
    max_epochs, accumulate_grad_batches, limit_train_batches, num_workers, num_samples, batch_size, drop_last
):
    """Compute effective max_steps from the provided parameters.

    Parameters
    ----------
    max_epochs : int
        Maximum number of epochs to train for.
    accumulate_grad_batches : int
        Number of batches to accumulate gradients for.
    limit_train_batches : int
        Number of batches to train for.
    num_workers : int
        Number of workers to use for training.
    num_samples : int
        Number of samples in the dataset.
    batch_size : int
        Batch size.
    drop_last : bool
        Whether to drop the last batch or not.
    """
    _round = math.floor if drop_last else math.ceil

    sampler_num_samples = math.ceil(num_samples / max(1, num_workers))

    if drop_last and num_workers > 1:
        logging.warning(
            "Please note that drop_last is broken in pytorch 1.6.0. We will fix when pytorch 1.7.0 is released"
        )
        # TODO: Master version, not in pytorch 1.6.0

    steps_per_epoch = _round(sampler_num_samples / batch_size)
    if isinstance(limit_train_batches, int) or limit_train_batches == 0.0:
        steps_per_epoch = min(steps_per_epoch, int(limit_train_batches))
    elif steps_per_epoch != float("inf"):
        # limit_train_batches is a percentage of batches per epoch
        steps_per_epoch = int(steps_per_epoch * limit_train_batches)

    return math.ceil(steps_per_epoch / accumulate_grad_batches) * max_epochs


AVAILABLE_SCHEDULERS = {
    "WarmupPolicy": WarmupPolicy,
    "WarmupHoldPolicy": WarmupHoldPolicy,
    "SquareAnnealing": SquareAnnealing,
    "CosineAnnealing": CosineAnnealing,
    "NoamAnnealing": NoamAnnealing,
    "NoamHoldAnnealing": NoamHoldAnnealing,
    "WarmupAnnealing": WarmupAnnealing,
    "InverseSquareRootAnnealing": InverseSquareRootAnnealing,
    "T5InverseSquareRootAnnealing": T5InverseSquareRootAnnealing,
    "SquareRootAnnealing": SquareRootAnnealing,
    "PolynomialDecayAnnealing": PolynomialDecayAnnealing,
    "PolynomialHoldDecayAnnealing": PolynomialHoldDecayAnnealing,
    "StepLR": pt_scheduler.StepLR,
    "ExponentialLR": pt_scheduler.ExponentialLR,
    "ReduceLROnPlateau": pt_scheduler.ReduceLROnPlateau,
    "CyclicLR": pt_scheduler.CyclicLR,
}

EPOCH_SCHEDULERS = {
    "ExponentialLR": pt_scheduler.ExponentialLR,
    "ReduceLROnPlateau": pt_scheduler.ReduceLROnPlateau,
}



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import os
import warnings
from abc import ABC
from collections import defaultdict
from pathlib import Path
from typing import Dict, Tuple, Union

import h5py
import numpy as np
import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer
from torch.utils.data import DataLoader

from atommic.collections.common.data.subsample import create_masker
from atommic.collections.common.losses import VALID_SEGMENTATION_LOSSES
from atommic.collections.common.losses.aggregator import AggregatorLoss
from atommic.collections.common.nn.base import BaseMRIModel, DistributedMetricSum
from atommic.collections.common.parts.utils import complex_abs, complex_abs_sq, is_none, unnormalize
from atommic.collections.segmentation.data.mri_segmentation_loader import (
    BraTS2023AdultGliomaSegmentationMRIDataset,
    ISLES2022SubAcuteStrokeSegmentationMRIDataset,
    SegmentationMRIDataset,
    SKMTEASegmentationMRIDataset,
)
from atommic.collections.segmentation.losses.cross_entropy import CrossEntropyLoss
from atommic.collections.segmentation.losses.dice import Dice
from atommic.collections.segmentation.parts.transforms import SegmentationMRIDataTransforms

__all__ = ["BaseMRISegmentationModel"]


class BaseMRISegmentationModel(BaseMRIModel, ABC):
    """Base class of all MRI Segmentation models."""

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`BaseMRISegmentationModel`.

        Parameters
        ----------
        cfg: DictConfig
            The configuration file.
        trainer: Trainer
            The PyTorch Lightning trainer.
        """
        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.acc = 1  # fixed acceleration factor to ensure acc is not None

        # Initialize the dimensionality of the data. It can be 2D or 2.5D -> meaning 2D with > 1 slices or 3D.
        self.dimensionality = cfg_dict.get("dimensionality", 2)
        self.consecutive_slices = cfg_dict.get("consecutive_slices", 1)

        # Set input channels.
        self.input_channels = cfg_dict.get("segmentation_module_input_channels", 2)
        if self.input_channels == 0:
            raise ValueError("Segmentation module input channels cannot be 0.")

        # Set type of data, i.e., magnitude only or complex valued.
        self.magnitude_input = cfg_dict.get("magnitude_input", True)
        # Refers to the type of the complex-valued data. It can be either "stacked" or "complex_abs" or
        # "complex_sqrt_abs".
        self.complex_valued_type = cfg_dict.get("complex_valued_type", "stacked")

        # Set normalization parameters for logging
        self.unnormalize_loss_inputs = cfg_dict.get("unnormalize_loss_inputs", False)
        self.unnormalize_log_outputs = cfg_dict.get("unnormalize_log_outputs", False)
        self.normalization_type = cfg_dict.get("normalization_type", "max")
        self.normalize_segmentation_output = cfg_dict.get("normalize_segmentation_output", True)

        # Whether to log multiple modalities, e.g. T1, T2, and FLAIR will be stacked and logged.
        self.log_multiple_modalities = cfg_dict.get("log_multiple_modalities", False)

        # Set threshold for segmentation classes. If None, no thresholding is applied.
        self.segmentation_classes_thresholds = cfg_dict.get("segmentation_classes_thresholds", None)
        self.segmentation_activation = cfg_dict.get("segmentation_activation", None)

        # Initialize loss related parameters.
        self.segmentation_losses = {}
        segmentation_loss = cfg_dict.get("segmentation_loss")
        segmentation_losses_ = {}
        if segmentation_loss is not None:
            for k, v in segmentation_loss.items():
                if k not in VALID_SEGMENTATION_LOSSES:
                    raise ValueError(
                        f"Segmentation loss {k} is not supported. Please choose one of the following: "
                        f"{VALID_SEGMENTATION_LOSSES}."
                    )
                if v is None or v == 0.0:
                    warnings.warn(f"The weight of segmentation loss {k} is set to 0.0. This loss will not be used.")
                else:
                    segmentation_losses_[k] = v
        else:
            # Default segmentation loss is Dice.
            segmentation_losses_["dice"] = 1.0
        if sum(segmentation_losses_.values()) != 1.0:
            warnings.warn("Sum of segmentation losses weights is not 1.0. Adjusting weights to sum up to 1.0.")
            total_weight = sum(segmentation_losses_.values())
            segmentation_losses_ = {k: v / total_weight for k, v in segmentation_losses_.items()}
        for name in VALID_SEGMENTATION_LOSSES:
            if name in segmentation_losses_:
                if name == "cross_entropy":
                    cross_entropy_loss_classes_weight = torch.tensor(
                        cfg_dict.get("cross_entropy_loss_classes_weight", 0.5)
                    )
                    self.segmentation_losses[name] = CrossEntropyLoss(
                        num_samples=cfg_dict.get("cross_entropy_loss_num_samples", 50),
                        ignore_index=cfg_dict.get("cross_entropy_loss_ignore_index", -100),
                        reduction=cfg_dict.get("cross_entropy_loss_reduction", "none"),
                        label_smoothing=cfg_dict.get("cross_entropy_loss_label_smoothing", 0.0),
                        weight=cross_entropy_loss_classes_weight,
                    )
                elif name == "dice":
                    self.segmentation_losses[name] = Dice(
                        include_background=cfg_dict.get("dice_loss_include_background", False),
                        to_onehot_y=cfg_dict.get("dice_loss_to_onehot_y", False),
                        sigmoid=cfg_dict.get("dice_loss_sigmoid", True),
                        softmax=cfg_dict.get("dice_loss_softmax", False),
                        other_act=cfg_dict.get("dice_loss_other_act", None),
                        squared_pred=cfg_dict.get("dice_loss_squared_pred", False),
                        jaccard=cfg_dict.get("dice_loss_jaccard", False),
                        flatten=cfg_dict.get("dice_loss_flatten", False),
                        reduction=cfg_dict.get("dice_loss_reduction", "mean"),
                        smooth_nr=cfg_dict.get("dice_loss_smooth_nr", 1e-5),
                        smooth_dr=cfg_dict.get("dice_loss_smooth_dr", 1e-5),
                        batch=cfg_dict.get("dice_loss_batch", False),
                    )
        self.segmentation_losses = {f"loss_{i+1}": v for i, v in enumerate(self.segmentation_losses.values())}
        self.total_segmentation_losses = len(self.segmentation_losses)
        self.total_segmentation_loss_weight = cfg_dict.get("total_segmentation_loss_weight", 1.0)

        # Set the metrics
        cross_entropy_metric_num_samples = cfg_dict.get("cross_entropy_metric_num_samples", 50)
        cross_entropy_metric_ignore_index = cfg_dict.get("cross_entropy_metric_ignore_index", -100)
        cross_entropy_metric_reduction = cfg_dict.get("cross_entropy_metric_reduction", "none")
        cross_entropy_metric_label_smoothing = cfg_dict.get("cross_entropy_metric_label_smoothing", 0.0)
        cross_entropy_metric_classes_weight = cfg_dict.get("cross_entropy_metric_classes_weight", None)
        dice_metric_include_background = cfg_dict.get("dice_metric_include_background", False)
        dice_metric_to_onehot_y = cfg_dict.get("dice_metric_to_onehot_y", False)
        dice_metric_sigmoid = cfg_dict.get("dice_metric_sigmoid", True)
        dice_metric_softmax = cfg_dict.get("dice_metric_softmax", False)
        dice_metric_other_act = cfg_dict.get("dice_metric_other_act", None)
        dice_metric_squared_pred = cfg_dict.get("dice_metric_squared_pred", False)
        dice_metric_jaccard = cfg_dict.get("dice_metric_jaccard", False)
        dice_metric_flatten = cfg_dict.get("dice_metric_flatten", False)
        dice_metric_reduction = cfg_dict.get("dice_metric_reduction", "mean")
        dice_metric_smooth_nr = cfg_dict.get("dice_metric_smooth_nr", 1e-5)
        dice_metric_smooth_dr = cfg_dict.get("dice_metric_smooth_dr", 1e-5)
        dice_metric_batch = cfg_dict.get("dice_metric_batch", True)

        # Initialize the module
        super().__init__(cfg=cfg, trainer=trainer)

        if not is_none(cross_entropy_metric_classes_weight):
            cross_entropy_metric_classes_weight = torch.tensor(cross_entropy_metric_classes_weight)
            self.cross_entropy_metric = CrossEntropyLoss(
                num_samples=cross_entropy_metric_num_samples,
                ignore_index=cross_entropy_metric_ignore_index,
                reduction=cross_entropy_metric_reduction,
                label_smoothing=cross_entropy_metric_label_smoothing,
                weight=cross_entropy_metric_classes_weight,
            )
        else:
            self.cross_entropy_metric = None  # type: ignore
        self.dice_metric = Dice(
            include_background=dice_metric_include_background,
            to_onehot_y=dice_metric_to_onehot_y,
            sigmoid=dice_metric_sigmoid,
            softmax=dice_metric_softmax,
            other_act=dice_metric_other_act,
            squared_pred=dice_metric_squared_pred,
            jaccard=dice_metric_jaccard,
            flatten=dice_metric_flatten,
            reduction=dice_metric_reduction,
            smooth_nr=dice_metric_smooth_nr,
            smooth_dr=dice_metric_smooth_dr,
            batch=dice_metric_batch,
        )

        # Set aggregation loss
        self.total_segmentation_loss = AggregatorLoss(
            num_inputs=self.total_segmentation_losses, weights=list(segmentation_losses_.values())
        )

        # Set distributed metrics
        self.CROSS_ENTROPY = DistributedMetricSum()
        self.DICE = DistributedMetricSum()
        self.cross_entropy_vals: Dict = defaultdict(dict)
        self.dice_vals: Dict = defaultdict(dict)
        self.TotExamples = DistributedMetricSum()

    def __abs_output__(self, x: torch.Tensor) -> torch.Tensor:
        """Converts the input to absolute value."""
        if x.shape[-1] == 2 or torch.is_complex(x):
            if self.complex_valued_type == "stacked":
                if x.shape[-1] == 2:
                    x = torch.view_as_complex(x)
            elif self.complex_valued_type == "complex_abs":
                x = complex_abs(x)
            elif self.complex_valued_type == "complex_sqrt_abs":
                x = complex_abs_sq(x)
        return x

    def __unnormalize_for_loss_or_log__(
        self,
        target: torch.Tensor,
        prediction: torch.Tensor,
        attrs: Dict,
        batch_idx: int = 0,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Unnormalizes the data for computing the loss or logging.

        Parameters
        ----------
        target : torch.Tensor
            Target data of shape [batch_size, n_x, n_y, 2].
        prediction : torch.Tensor
            Prediction data of shape [batch_size, n_x, n_y, 2].
        attrs : Dict
            Attributes of the data with pre normalization values.
        batch_idx : int
            Batch index. Default is ``0``.

        Returns
        -------
        target : torch.Tensor
            Unnormalized target data.
        prediction : torch.Tensor
            Unnormalized prediction data.
        sensitivity_maps : torch.Tensor
            Unnormalized sensitivity maps.
        """
        target = unnormalize(
            target,
            {
                "min": attrs["target_min"][batch_idx],
                "max": attrs["target_max"][batch_idx],
                "mean": attrs["target_mean"][batch_idx],
                "std": attrs["target_std"][batch_idx],
            },
            self.normalization_type,
        )
        prediction = unnormalize(
            prediction,
            {
                "min": attrs["prediction_min"][batch_idx],
                "max": attrs["prediction_max"][batch_idx],
                "mean": attrs["prediction_mean"][batch_idx],
                "std": attrs["prediction_std"][batch_idx],
            },
            self.normalization_type,
        )

        return target, prediction

    def process_segmentation_loss(self, target: torch.Tensor, prediction: torch.Tensor, attrs: Dict) -> Dict:
        """Processes the segmentation loss.

        Parameters
        ----------
        target : torch.Tensor
            Target data of shape [batch_size, nr_classes, n_x, n_y].
        prediction : torch.Tensor
            Prediction of shape [batch_size, nr_classes, n_x, n_y].
        attrs : Dict
            Attributes of the data with pre normalization values.

        Returns
        -------
        Dict
            Dictionary containing the (multiple) loss values. For example, if the cross entropy loss and the dice loss
            are used, the dictionary will contain the keys ``cross_entropy_loss``, ``dice_loss``, and
            (combined) ``segmentation_loss``.
        """
        if self.unnormalize_loss_inputs:
            target, prediction = self.__unnormalize_for_loss_or_log__(target, prediction, attrs)
        losses = {}
        for name, loss_func in self.segmentation_losses.items():
            loss = loss_func(target, prediction)
            if isinstance(loss, tuple):
                # In case of the dice loss, the loss is a tuple of the form (dice, dice loss)
                loss = loss[1]
            losses[name] = loss
        return self.total_segmentation_loss(**losses) * self.total_segmentation_loss_weight

    def __compute_and_log_metrics_and_outputs__(
        self,
        predictions: Union[list, torch.Tensor],
        target_reconstruction: torch.Tensor,
        target_segmentation: torch.Tensor,
        attrs: Dict,
        fname: str,
        slice_idx: int,
    ):
        """Computes the metrics and logs the outputs.

        Parameters
        ----------
        predictions : torch.Tensor
            Prediction data of shape [batch_size, n_x, n_y].
        target_reconstruction : torch.Tensor
            Target reconstruction data of shape [batch_size, n_x, n_y].
        target_segmentation : torch.Tensor
            Target segmentation data of shape [batch_size, segmentation_classes, n_x, n_y].
        attrs : Dict
            Attributes of the data with pre normalization values.
        fname : str
            File name.
        slice_idx : int
            Slice index.
        """
        if isinstance(predictions, list):
            while isinstance(predictions, list):
                predictions = predictions[-1]

        # Ensure all inputs are both viewed in the same way.
        target_reconstruction = self.__abs_output__(target_reconstruction)

        if self.consecutive_slices > 1:
            # reshape the target and prediction to [batch_size * consecutive_slices, nr_classes, n_x, n_y]
            batch_size = target_segmentation.shape[0] // self.consecutive_slices
            slice_to_keep = self.consecutive_slices // 2
            target_segmentation = target_segmentation.reshape(
                batch_size, self.consecutive_slices, *target_segmentation.shape[1:]
            )[:, slice_to_keep]
            predictions = predictions.reshape(batch_size, self.consecutive_slices, *predictions.shape[1:])[
                :, slice_to_keep
            ]
            target_reconstruction = target_reconstruction[:, slice_to_keep]
        else:
            batch_size = target_segmentation.shape[0]

        # Iterate over the batch and log the target and predictions.
        for _batch_idx_ in range(batch_size):
            output_target_segmentation = target_segmentation[_batch_idx_]
            output_predictions = predictions[_batch_idx_]

            if self.unnormalize_log_outputs:
                # Unnormalize target and predictions with pre normalization values. This is only for logging purposes.
                # For the loss computation, the self.unnormalize_loss_inputs flag is used.
                output_target_segmentation, output_predictions = self.__unnormalize_for_loss_or_log__(
                    output_target_segmentation, output_predictions, attrs, batch_idx=_batch_idx_
                )

            output_target_segmentation = output_target_segmentation.detach().cpu()
            output_predictions = output_predictions.detach().cpu()

            # Log target and predictions, if log_image is True for this slice.
            if attrs["log_image"][_batch_idx_]:
                key = f"{fname[_batch_idx_]}_slice_{int(slice_idx[_batch_idx_])}"  # type: ignore

                # Normalize (reconstruction) target to [0, 1] for logging.
                output_target_reconstruction = torch.abs(target_reconstruction[_batch_idx_]).float()
                output_target_reconstruction = output_target_reconstruction / torch.max(output_target_reconstruction)

                if self.log_multiple_modalities:
                    # concatenate the reconstruction predictions for logging
                    output_target_reconstruction = torch.cat(
                        [output_target_reconstruction[i] for i in range(output_target_reconstruction.shape[0])], dim=-1
                    )

                self.log_image(f"{key}/a/input", output_target_reconstruction)

                # concatenate the segmentation classes for logging
                target_segmentation_classes = torch.cat(
                    [output_target_segmentation[i] for i in range(output_target_segmentation.shape[0])], dim=-1
                )
                output_predictions_segmentation_classes = torch.cat(
                    [output_predictions[i] for i in range(output_predictions.shape[0])], dim=-1
                )
                self.log_image(f"{key}/b/segmentation_labels", target_segmentation_classes)
                self.log_image(f"{key}/c/segmentation_predictions", output_predictions_segmentation_classes)
                self.log_image(
                    f"{key}/d/segmentation_error",
                    torch.abs(target_segmentation_classes - output_predictions_segmentation_classes),
                )

            output_target_segmentation = output_target_segmentation.unsqueeze(0)
            output_predictions = output_predictions.unsqueeze(0)

            if self.cross_entropy_metric is not None:
                self.cross_entropy_vals[fname][slice_idx] = self.cross_entropy_metric(
                    output_target_segmentation.to(self.device), output_predictions.to(self.device)
                )

            dice_score, _ = self.dice_metric(output_target_segmentation, output_predictions)
            self.dice_vals[fname][slice_idx] = dice_score

    def inference_step(
        self, image: torch.Tensor, target: torch.Tensor, fname: str, slice_idx: int, attrs: Dict
    ) -> Dict[str, torch.Tensor]:
        """Performs an inference step, i.e., computes the predictions of the model.

        Parameters
        ----------
        image : torch.Tensor
            Input data. Shape [batch_size, n_x, n_y, 2].
        target : torch.Tensor
            Target data. Shape [batch_size, n_x, n_y, 2].
        fname : str
            File name.
        slice_idx : int
            Slice index.
        attrs : Dict
            Attributes dictionary.

        Returns
        -------
        Dict[str, torch.Tensor]
            Dictionary of processed inputs and model's predictions, with keys:
                'fname' : str
                    File name.
                'slice_idx' : int
                    Slice index.
                'predictions' : Union[List[torch.Tensor], torch.Tensor]
                    Model's predictions. Shape [batch_size, segmentation_classes, n_x, n_y, 2].
                'target' : torch.Tensor
                    Target data. Shape [batch_size, n_x, n_y, 2].
                'attrs' : dict
                    Attributes dictionary.
        """
        # Model forward pass
        prediction = self.forward(image)

        if self.consecutive_slices > 1:
            # reshape the target and prediction to [batch_size * consecutive_slices, nr_classes, n_x, n_y]
            batch_size, slices = prediction.shape[:2]
            if target.dim() == 5:
                target = target.reshape(batch_size * slices, *target.shape[2:])
            if prediction.dim() == 5:
                prediction = prediction.reshape(batch_size * slices, *prediction.shape[2:])

        if not is_none(self.segmentation_classes_thresholds):
            for class_idx, thres in enumerate(self.segmentation_classes_thresholds):
                if self.segmentation_activation == "sigmoid":
                    cond = torch.sigmoid(prediction[:, class_idx])
                elif self.segmentation_activation == "softmax":
                    cond = torch.softmax(prediction[:, class_idx], dim=1)
                else:
                    cond = prediction[:, class_idx]
                prediction[:, class_idx] = torch.where(
                    cond >= thres, prediction[:, class_idx], torch.zeros_like(prediction[:, class_idx])
                )

        return {
            "fname": fname,
            "slice_idx": slice_idx,
            "predictions": prediction,
            "target": target,
            "attrs": attrs,
        }

    def training_step(self, batch: Dict[float, torch.Tensor], batch_idx: int) -> Dict[str, torch.Tensor]:
        """Performs a training step.

        Parameters
        ----------
        batch : Dict[float, torch.Tensor]
            Batch of data with keys:
                'kspace' : List of torch.Tensor
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'y' : Union[torch.Tensor, None]
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'sensitivity_maps' : torch.Tensor
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'mask' : Union[torch.Tensor, None]
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'initial_prediction_reconstruction' : torch.Tensor
                    Initial reconstruction prediction. Shape [batch_size, n_x, n_y, 2].
                'target_reconstruction' : torch.Tensor
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'segmentation_labels' : torch.Tensor
                    Target segmentation labels. Shape [batch_size, segmentation_classes, n_x, n_y].
                'fname' : str
                    File name.
                'slice_idx' : int
                    Slice index.
                'acceleration' : float
                    Acceleration factor of the sampling mask.
                'attrs' : dict
                    Attributes dictionary.

        batch_idx : int
            Batch index.

        Returns
        -------
        Dict[str, torch.Tensor]
            Dictionary of loss and log.
        """
        (
            _,
            _,
            _,
            _,
            initial_reconstruction_prediction,
            _,
            target_segmentation,
            fname,
            slice_idx,
            _,
            attrs,
        ) = batch

        # In case of complex (fully-sampled) data the initial_reconstruction_prediction is a list of tensors of len 1.
        if isinstance(initial_reconstruction_prediction, list):
            initial_reconstruction_prediction = initial_reconstruction_prediction[-1]

        outputs = self.inference_step(
            initial_reconstruction_prediction,
            target_segmentation,
            fname,  # type: ignore
            slice_idx,  # type: ignore
            attrs,  # type: ignore
        )

        train_loss = self.process_segmentation_loss(outputs["target"], outputs["predictions"], attrs)  # type: ignore

        tensorboard_logs = {
            "train_loss": train_loss.item(),  # type: ignore
            "lr": self._optimizer.param_groups[0]["lr"],  # type: ignore
        }

        self.log(
            "train_segmentation_loss",
            train_loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            batch_size=1,  # type: ignore
            sync_dist=True,
        )

        return {"loss": train_loss, "log": tensorboard_logs}

    def validation_step(self, batch: Dict[float, torch.Tensor], batch_idx: int):
        """Performs a validation step.

        Parameters
        ----------
        batch : Dict[float, torch.Tensor]
            Batch of data with keys:
                'kspace' : List of torch.Tensor
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'y' : Union[torch.Tensor, None]
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'sensitivity_maps' : torch.Tensor
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'mask' : Union[torch.Tensor, None]
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'initial_prediction_reconstruction' : torch.Tensor
                    Initial reconstruction prediction. Shape [batch_size, n_x, n_y, 2].
                'target_reconstruction' : torch.Tensor
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'segmentation_labels' : torch.Tensor
                    Target segmentation labels. Shape [batch_size, segmentation_classes, n_x, n_y].
                'fname' : str
                    File name.
                'slice_idx' : int
                    Slice index.
                'acceleration' : float
                    Acceleration factor of the sampling mask.
                'attrs' : dict
                    Attributes dictionary.

        batch_idx : int
            Batch index.
        """
        (
            _,
            _,
            _,
            _,
            initial_reconstruction_prediction,
            target_reconstruction,
            target_segmentation,
            fname,
            slice_idx,
            _,
            attrs,
        ) = batch

        # In case of complex (fully-sampled) data the initial_reconstruction_prediction is a list of tensors of len 1.
        if isinstance(initial_reconstruction_prediction, list):
            initial_reconstruction_prediction = initial_reconstruction_prediction[-1]

        outputs = self.inference_step(
            initial_reconstruction_prediction,
            target_segmentation,
            fname,  # type: ignore
            slice_idx,  # type: ignore
            attrs,  # type: ignore
        )

        target_segmentation = outputs["target"]
        predictions = outputs["predictions"]

        # print memory usage for debugging
        val_loss = self.process_segmentation_loss(target_segmentation, predictions, attrs)  # type: ignore

        # Compute metrics and log them and log outputs.
        self.__compute_and_log_metrics_and_outputs__(
            predictions,
            target_reconstruction,
            target_segmentation,
            attrs,  # type: ignore
            fname,  # type: ignore
            slice_idx,  # type: ignore
        )

        self.validation_step_outputs.append({"val_loss": val_loss})

    def test_step(self, batch: Dict[float, torch.Tensor], batch_idx: int):
        """Performs a test step.

        Parameters
        ----------
        batch : Dict[float, torch.Tensor]
            Batch of data with keys:
                'kspace' : List of torch.Tensor
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'y' : Union[torch.Tensor, None]
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'sensitivity_maps' : torch.Tensor
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'mask' : Union[torch.Tensor, None]
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'initial_prediction_reconstruction' : torch.Tensor
                    Initial reconstruction prediction. Shape [batch_size, n_x, n_y, 2].
                'target_reconstruction' : torch.Tensor
                    Placeholder to keep the same structure as the base (task) classes. Not used.
                'segmentation_labels' : torch.Tensor
                    Target segmentation labels. Shape [batch_size, segmentation_classes, n_x, n_y].
                'fname' : str
                    File name.
                'slice_idx' : int
                    Slice index.
                'acceleration' : float
                    Acceleration factor of the sampling mask.
                'attrs' : dict
                    Attributes dictionary.

        batch_idx : int
            Batch index.
        """
        (
            _,
            _,
            _,
            _,
            initial_reconstruction_prediction,
            target_reconstruction,
            target_segmentation,
            fname,
            slice_idx,
            _,
            attrs,
        ) = batch

        # In case of complex (fully-sampled) data the initial_reconstruction_prediction is a list of tensors of len 1.
        if isinstance(initial_reconstruction_prediction, list):
            initial_reconstruction_prediction = initial_reconstruction_prediction[-1]

        outputs = self.inference_step(
            initial_reconstruction_prediction,
            target_segmentation,
            fname,  # type: ignore
            slice_idx,  # type: ignore
            attrs,  # type: ignore
        )

        target_segmentation = outputs["target"]
        predictions = outputs["predictions"]

        # Compute metrics and log them and log outputs.
        self.__compute_and_log_metrics_and_outputs__(
            predictions,
            target_reconstruction,
            target_segmentation,
            attrs,  # type: ignore
            fname,  # type: ignore
            slice_idx,  # type: ignore
        )

        # Get the file name.
        fname = attrs['fname'][0]  # type: ignore
        if '.nii.gz' in fname or '.nii' in fname or '.h5' in fname:  # type: ignore
            fname = fname.split('.')[0]  # type: ignore

        self.test_step_outputs.append([fname, slice_idx, predictions.detach().cpu()])

    def on_validation_epoch_end(self):
        """Called at the end of validation epoch to aggregate outputs.

        Returns
        -------
        metrics : dict
            Dictionary of metrics.
        """
        self.log("val_loss", torch.stack([x["val_loss"] for x in self.validation_step_outputs]).mean(), sync_dist=True)

        # Log metrics.
        if self.cross_entropy_metric is not None:
            cross_entropy_vals = defaultdict(dict)
            for k, v in self.cross_entropy_vals.items():
                cross_entropy_vals[k].update(v)

        dice_vals = defaultdict(dict)
        for k, v in self.dice_vals.items():
            dice_vals[k].update(v)

        metrics = {"Cross_Entropy": 0, "DICE": 0}

        local_examples = 0
        for fname in dice_vals:
            local_examples += 1
            if self.cross_entropy_metric is not None:
                metrics["Cross_Entropy"] = metrics["Cross_Entropy"] + torch.mean(
                    torch.cat([v.view(-1) for _, v in cross_entropy_vals[fname].items()])
                )
            metrics["DICE"] = metrics["DICE"] + torch.mean(
                torch.cat([v.view(-1) for _, v in dice_vals[fname].items()])
            )

        # reduce across ddp via sum
        if self.cross_entropy_metric is not None:
            metrics["Cross_Entropy"] = self.CROSS_ENTROPY(metrics["Cross_Entropy"])
        metrics["DICE"] = self.DICE(metrics["DICE"])
        tot_examples = self.TotExamples(torch.tensor(local_examples))

        for metric, value in metrics.items():
            self.log(f"val_metrics/{metric}", value / tot_examples, prog_bar=True, sync_dist=True)

    def on_test_epoch_end(self):  # noqa: MC0001
        """Called at the end of test epoch to aggregate outputs, log metrics and save predictions.

        Returns
        -------
        metrics : dict
            Dictionary of metrics.
        """
        # Log metrics.
        if self.cross_entropy_metric is not None:
            cross_entropy_vals = defaultdict(dict)
            for k, v in self.cross_entropy_vals.items():
                cross_entropy_vals[k].update(v)

        dice_vals = defaultdict(dict)
        for k, v in self.dice_vals.items():
            dice_vals[k].update(v)

        metrics = {"Cross_Entropy": 0, "DICE": 0}

        local_examples = 0
        for fname in dice_vals:
            local_examples += 1
            if self.cross_entropy_metric is not None:
                metrics["Cross_Entropy"] = metrics["Cross_Entropy"] + torch.mean(
                    torch.cat([v.view(-1) for _, v in cross_entropy_vals[fname].items()])
                )
            metrics["DICE"] = metrics["DICE"] + torch.mean(
                torch.cat([v.view(-1) for _, v in dice_vals[fname].items()])
            )

        # reduce across ddp via sum
        if self.cross_entropy_metric is not None:
            metrics["Cross_Entropy"] = self.CROSS_ENTROPY(metrics["Cross_Entropy"])
        metrics["DICE"] = self.DICE(metrics["DICE"])
        tot_examples = self.TotExamples(torch.tensor(local_examples))

        for metric, value in metrics.items():
            self.log(f"test_metrics/{metric}", value / tot_examples, prog_bar=True, sync_dist=True)

        segmentations = defaultdict(list)
        for fname, slice_num, segmentations_pred in self.test_step_outputs:
            segmentations[fname].append((slice_num, segmentations_pred))

        for fname in segmentations:
            segmentations[fname] = np.stack([out for _, out in sorted(segmentations[fname])])

        if self.consecutive_slices > 1:
            # iterate over the slices and always keep the middle slice
            for fname in segmentations:
                segmentations[fname] = segmentations[fname][:, self.consecutive_slices // 2]

        if "wandb" in self.logger.__module__.lower():
            out_dir = Path(os.path.join(self.logger.save_dir, "segmentations"))
        else:
            out_dir = Path(os.path.join(self.logger.log_dir, "segmentations"))
        out_dir.mkdir(exist_ok=True, parents=True)

        for fname, segmentations_pred in segmentations.items():
            with h5py.File(out_dir / fname, "w") as hf:
                hf.create_dataset("segmentation", data=segmentations_pred)

    @staticmethod
    def _setup_dataloader_from_config(cfg: DictConfig) -> DataLoader:
        """Setups the dataloader from the configuration (yaml) file.

        Parameters
        ----------
        cfg : DictConfig
            Configuration file.

        Returns
        -------
        dataloader : torch.utils.data.DataLoader
            Dataloader.
        """
        mask_root = cfg.get("mask_path", None)
        mask_args = cfg.get("mask_args", None)
        shift_mask = mask_args.get("shift_mask", False)
        mask_type = mask_args.get("type", None)

        mask_func = None
        mask_center_scale = 0.02

        if is_none(mask_root) and not is_none(mask_type):
            accelerations = mask_args.get("accelerations", [1])
            accelerations = list(accelerations)
            if len(accelerations) == 1:
                accelerations = accelerations * 2
            center_fractions = mask_args.get("center_fractions", [1])
            center_fractions = list(center_fractions)
            if len(center_fractions) == 1:
                center_fractions = center_fractions * 2
            mask_center_scale = mask_args.get("center_scale", 0.02)

            mask_func = [create_masker(mask_type, center_fractions, accelerations)]

        complex_data = cfg.get("complex_data", True)

        dataset_format = cfg.get("dataset_format", None)
        if dataset_format.lower() == "brats2023adultglioma":
            dataloader = BraTS2023AdultGliomaSegmentationMRIDataset
        elif dataset_format.lower() == "isles2022subacutestroke":
            dataloader = ISLES2022SubAcuteStrokeSegmentationMRIDataset
        elif dataset_format.lower() in (
            "skm-tea-echo1",
            "skm-tea-echo2",
            "skm-tea-echo1+echo2",
            "skm-tea-echo1+echo2-mc",
            "skm-tea-echo1+echo2-rss",
        ):
            dataloader = SKMTEASegmentationMRIDataset
        else:
            dataloader = SegmentationMRIDataset

        dataset = dataloader(
            root=cfg.get("data_path"),
            coil_sensitivity_maps_root=cfg.get("coil_sensitivity_maps_path", None),
            mask_root=mask_root,
            noise_root=cfg.get("noise_path", None),
            initial_predictions_root=cfg.get("initial_predictions_path"),
            dataset_format=dataset_format,
            sample_rate=cfg.get("sample_rate", 1.0),
            volume_sample_rate=cfg.get("volume_sample_rate", None),
            use_dataset_cache=cfg.get("use_dataset_cache", False),
            dataset_cache_file=cfg.get("dataset_cache_file", None),
            num_cols=cfg.get("num_cols", None),
            consecutive_slices=cfg.get("consecutive_slices", 1),
            data_saved_per_slice=cfg.get("data_saved_per_slice", False),
            n2r_supervised_rate=cfg.get("n2r_supervised_rate", 0.0),
            complex_target=cfg.get("complex_target", False),
            log_images_rate=cfg.get("log_images_rate", 1.0),
            transform=SegmentationMRIDataTransforms(
                complex_data=complex_data,
                dataset_format=dataset_format,
                apply_prewhitening=cfg.get("apply_prewhitening", False),
                find_patch_size=cfg.get("find_patch_size", False),
                prewhitening_scale_factor=cfg.get("prewhitening_scale_factor", 1.0),
                prewhitening_patch_start=cfg.get("prewhitening_patch_start", 10),
                prewhitening_patch_length=cfg.get("prewhitening_patch_length", 30),
                apply_gcc=cfg.get("apply_gcc", False),
                gcc_virtual_coils=cfg.get("gcc_virtual_coils", 10),
                gcc_calib_lines=cfg.get("gcc_calib_lines", 10),
                gcc_align_data=cfg.get("gcc_align_data", False),
                apply_random_motion=cfg.get("apply_random_motion", False),
                random_motion_type=cfg.get("random_motion_type", "gaussian"),
                random_motion_percentage=cfg.get("random_motion_percentage", [10, 10]),
                random_motion_angle=cfg.get("random_motion_angle", 10),
                random_motion_translation=cfg.get("random_motion_translation", 10),
                random_motion_center_percentage=cfg.get("random_motion_center_percentage", 0.02),
                random_motion_num_segments=cfg.get("random_motion_num_segments", 8),
                random_motion_random_num_segments=cfg.get("random_motion_random_num_segments", True),
                random_motion_non_uniform=cfg.get("random_motion_non_uniform", False),
                estimate_coil_sensitivity_maps=cfg.get("estimate_coil_sensitivity_maps", False),
                coil_sensitivity_maps_type=cfg.get("coil_sensitivity_maps_type", "espirit"),
                coil_sensitivity_maps_gaussian_sigma=cfg.get("coil_sensitivity_maps_gaussian_sigma", 0.0),
                coil_sensitivity_maps_espirit_threshold=cfg.get("coil_sensitivity_maps_espirit_threshold", 0.05),
                coil_sensitivity_maps_espirit_kernel_size=cfg.get("coil_sensitivity_maps_espirit_kernel_size", 6),
                coil_sensitivity_maps_espirit_crop=cfg.get("coil_sensitivity_maps_espirit_crop", 0.95),
                coil_sensitivity_maps_espirit_max_iters=cfg.get("coil_sensitivity_maps_espirit_max_iters", 30),
                coil_combination_method=cfg.get("coil_combination_method", "SENSE"),
                dimensionality=cfg.get("dimensionality", 2),
                mask_func=mask_func,
                shift_mask=shift_mask,
                mask_center_scale=mask_center_scale,
                remask=cfg.get("remask", False),
                ssdu=cfg.get("ssdu", False),
                ssdu_mask_type=cfg.get("ssdu_mask_type", "Gaussian"),
                ssdu_rho=cfg.get("ssdu_rho", 0.4),
                ssdu_acs_block_size=cfg.get("ssdu_acs_block_size", (4, 4)),
                ssdu_gaussian_std_scaling_factor=cfg.get("ssdu_gaussian_std_scaling_factor", 4.0),
                ssdu_outer_kspace_fraction=cfg.get("ssdu_outer_kspace_fraction", 0.0),
                ssdu_export_and_reuse_masks=cfg.get("ssdu_export_and_reuse_masks", False),
                n2r=cfg.get("n2r", False),
                n2r_supervised_rate=cfg.get("n2r_supervised_rate", 0.0),
                n2r_probability=cfg.get("n2r_probability", 0.5),
                n2r_std_devs=cfg.get("n2r_std_devs", (0.0, 0.0)),
                n2r_rhos=cfg.get("n2r_rhos", (0.4, 0.4)),
                n2r_use_mask=cfg.get("n2r_use_mask", True),
                unsupervised_masked_target=cfg.get("unsupervised_masked_target", False),
                crop_size=cfg.get("crop_size", None),
                kspace_crop=cfg.get("kspace_crop", False),
                crop_before_masking=cfg.get("crop_before_masking", False),
                kspace_zero_filling_size=cfg.get("kspace_zero_filling_size", None),
                normalize_inputs=cfg.get("normalize_inputs", True),
                normalization_type=cfg.get("normalization_type", "max"),
                kspace_normalization=cfg.get("kspace_normalization", False),
                fft_centered=cfg.get("fft_centered", False),
                fft_normalization=cfg.get("fft_normalization", "backward"),
                spatial_dims=cfg.get("spatial_dims", None),
                coil_dim=cfg.get("coil_dim", 1),
                consecutive_slices=cfg.get("consecutive_slices", 1),
                use_seed=cfg.get("use_seed", True),
            ),
            segmentations_root=cfg.get("segmentations_path"),
            segmentation_classes=cfg.get("segmentation_classes", 2),
            segmentation_classes_to_remove=cfg.get("segmentation_classes_to_remove", None),
            segmentation_classes_to_combine=cfg.get("segmentation_classes_to_combine", None),
            segmentation_classes_to_separate=cfg.get("segmentation_classes_to_separate", None),
            segmentation_classes_thresholds=cfg.get("segmentation_classes_thresholds", None),
            complex_data=complex_data,
        )
        if cfg.shuffle:
            sampler = torch.utils.data.RandomSampler(dataset)
        else:
            sampler = torch.utils.data.SequentialSampler(dataset)

        return torch.utils.data.DataLoader(
            dataset=dataset,
            batch_size=cfg.get("batch_size", 1),
            sampler=sampler,
            num_workers=cfg.get("num_workers", 4),
            pin_memory=cfg.get("pin_memory", False),
            drop_last=cfg.get("drop_last", False),
        )
